/*******************************************************************
Copyright (c) 2002-2003 The University of Maryland.

Permission is hereby granted, without written agreement and without
license or royalty fees, to use, copy, modify, and distribute this
software and its documentation for any purpose, provided that the above
copyright notice and the following two paragraphs appear in all copies
of this software.

IN NO EVENT SHALL THE UNIVERSITY OF MARYLAND BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
THE UNIVERSITY OF MARYLAND HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

THE UNIVERSITY OF MARYLAND SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
MARYLAND HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS.
********************************************************************/

// import exceptions
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.NoSuchElementException;

// import classes
import java.io.BufferedReader;
import java.io.FileReader;
import java.util.Iterator;
import java.util.StringTokenizer;

//////////////////////////////////////////////////////////////////////////
//// BenchmarkReader
/**
This file is for benchmark file reading. Benchmark data specified in
integers can be retrieved by this class. Another powerful feature is
that this class can skip any text description, like comments, and
extract lines of pure integers.

@author Mingyung Ko
@version $Id: BenchmarkReader.java,v 1.8 2003/04/21 19:08:56 myko Exp $
*/

public class BenchmarkReader {

    ///////////////////////////////////////////////////////////////////
    ////                        static methods                     ////

    /** Read a line of integers from the buffered I/O. Only lines with
     *  integers will be read.
     *  @param reader The buffered I/O.
     *  @return An array of integers.
     */
    public static int[] integerReadLine(BufferedReader reader) {

        boolean lineInvalid = true;
        String line = new String();
        int[] integers = new int[1];
        try {
            nextLine:
            while (lineInvalid) {
                line = reader.readLine();
                // End of file, insufficient data
                if (line == null) {
                    throw new NullPointerException();
                // Skip empty lines
                } else if (line.length() == 0) {
                    continue;
                // Extract integers from the line
                } else {
                    StringTokenizer tokenizer = new StringTokenizer(line);
                    integers = new int[tokenizer.countTokens()];
                    for (int i = 0; i < integers.length; i++) {
                        try {
                            integers[i] =
                                    Integer.parseInt(tokenizer.nextToken());
                        // Skip non-integer lines
                        } catch (NumberFormatException e) {
                            continue nextLine;
                        }
                    }
                    lineInvalid = false;
                }
            }
        } catch (Exception e) {
            throw new BenchmarkReadingException("File reading failure.");
        }
        return integers;
    }

    /** Read a conflict graph. Conflict graphs are used in
     *  data partitioning problems. {@link APGAN} and {@link GDPPO}
     *  are used to schedule the SDF graph and compute buffer size.
     *  @param reader The file I/O buffer to read data in.
     *  @return A conflict graph in the form of
     *          {@link synthesis.dif.csdf.sdf.mem.PartitionedGraph}.
     */
    public static PartitionedGraph conflictGraphReader(
            BufferedReader reader) {

        PartitionedGraph conflictGraph = new PartitionedGraph();
        int[] line;

        // Test whether a SDF graph is included by the first line.
        try {
            reader.mark(20);
            line = integerReadLine(reader);
            reader.reset();
        } catch (IOException e) {
            throw new BenchmarkReadingException();
        }

        // SDF graph specification is contained.
        if (line.length == 2) {
            // Compute buffer size.
            _sdfgraph = SDFGraphReader(reader);
            SDFScheduler scheduler = new MultirateCycleBreaking(_sdfgraph);
            int[] bufferSize = scheduler.bufferSize();
            // Add buffers as nodes to the conflict graph.
            Iterator buffers = _sdfgraph.edges().iterator();
            while (buffers.hasNext()) {
                Edge buffer = (Edge)buffers.next();
                Double nodeValue = new Double(
                        (double)bufferSize[_sdfgraph.edgeLabel(buffer)]);
                conflictGraph.addNode(new Node(nodeValue));
            }
        }

        // No matter an SDF is given or not, state variables will be read
        // by all means.
        line = integerReadLine(reader);
        int stateVariableCounts = line[0];
        for (int i = 0; i < stateVariableCounts; i++) {
            line = integerReadLine(reader);
            conflictGraph.addNode(new Node(new Double((double)line[0])));
        }
        line = integerReadLine(reader);
        int conflictCounts = line[0];
        for (int i = 0; i < conflictCounts; i++) {
            line = integerReadLine(reader);
            Node n1 = conflictGraph.node(line[0]);
            Node n2 = conflictGraph.node(line[1]);
            conflictGraph.addEdge(new Edge(n1, n2));
        }

        return conflictGraph;
    }

    /** Read conflict graph from the given file.
     *  @param fileName The given file name.
     *  @param scheduler The scheduler to compute buffer size.
     *  @return The conflict graph.
     */
    public static PartitionedGraph conflictGraphReader(String fileName) {
        try {
            return conflictGraphReader(
                    new BufferedReader(new FileReader(fileName)));
        } catch (FileNotFoundException e) {
            throw new BenchmarkReadingException(
                    "File not found: " + fileName);
        }
    }

    /** Get the SDF graph associated with conflict graph reading.
     *  For conflict graph reading, please see
     *  {@link #conflictGraphReader(BufferedReader)}.
     *  @return The associated SDF graph.
     */
    public static SDFGraph getAssociatedSDFGraph() {
        return _sdfgraph;
    }

    /** Extract SDF graph from a <code>BufferedReader</code>.
     *  @param reader The file I/O buffer to read data in.
     *  @return The {@link SDFGraph}.
     */
    public static SDFGraph SDFGraphReader(BufferedReader reader) {
        SDFGraph graph = new SDFGraph();

        int   nodeCount, edgeCount;
        int[] line;

        // get node and edge counts
        line = integerReadLine(reader);
        if (line.length != 2) {
            throw new BenchmarkSpecException(
                    "Node and edge counts are both required.");
        } else {
            nodeCount = line[0];
            edgeCount = line[1];
            if (nodeCount < 1 | edgeCount < 1)
                throw new BenchmarkSpecException(
                        "Positive counts are required.");
        }
        // Add nodes in SDF graph
        Node[] nodes = new Node[nodeCount];
        for (int i = 0; i < nodeCount; i++) {
            String nodeName = new String() + 'a' + i;
            nodes[i] = graph.addNodeWeight(
                    new SDFNodeWeight(nodeName));
            try {
                graph.setName(nodes[i], nodeName);
            } catch (Exception e) {}
        }
        // Add edges in SDF graph
        for (int i = 0; i < edgeCount; i++) {
            // line: {source, sink, production, consumption, delay}
            line = integerReadLine(reader);
            // Error checking
            if (line.length < 2)
                throw new BenchmarkSpecException(
                        "Source or sink node missed.");
            if (line.length > 5)
                throw new BenchmarkSpecException(
                        "Edge elements over specified.");
            for (int j = 0; j < line.length; j++) {
                switch (j) {
                    case 0:
                    case 1:
                        if ((line[j] < 0) | (line[j] >= nodeCount))
                            throw new BenchmarkSpecException(
                                    "Invalid node label.");
                        break;
                    case 2:
                    case 3:
                        if (line[j] < 1)
                            throw new BenchmarkSpecException(
                                    "Positive token rate required.");
                        break;
                    case 4:
                        if (line[4] < 0)
                            throw new BenchmarkSpecException(
                                    "Negative delays disallowed.");
                }
            }
            // Set {source, sink, production, consumption, delay}
            SDFEdgeWeight edgeWeight = null;
            switch (line.length) {
                case 2:
                    edgeWeight = new SDFEdgeWeight(1, 1, 0);
                    break;
                case 3:
                    edgeWeight = new SDFEdgeWeight(line[2], line[2], 0);
                    break;
                case 4:
                    edgeWeight = new SDFEdgeWeight(line[2], line[3], 0);
                    break;
                default:
                    edgeWeight =
                            new SDFEdgeWeight(line[2], line[3], line[4]);
            }
            Node source = nodes[line[0]];
            Node sink   = nodes[line[1]];
            graph.addEdge(source, sink, edgeWeight);
        }

        return graph;
    }

    /** Extract SDF graph from a dfsched file.
     *  @param fileName The given dfsched file name.
     *  @return The {@link SDFGraph}.
     */
    public static SDFGraph SDFGraphReader(String fileName) {
        try {
            return SDFGraphReader(
                    new BufferedReader(new FileReader(fileName)));
        } catch (FileNotFoundException e) {
            throw new BenchmarkReadingException(
                    "File not found: " + fileName);
        }
    }

    ///////////////////////////////////////////////////////////////////
    ////                       static variables                    ////

    // The SDF graph associated with conflict graph reading.
    private static SDFGraph _sdfgraph;
}
