/* Information associated with an CSDF edge.

 Copyright (c) 2002 The University of Maryland  
 All rights reserved.
 Permission is hereby granted, without written agreement and without
 license or royalty fees, to use, copy, modify, and distribute this
 software and its documentation for any purpose, provided that the above
 copyright notice and the following two paragraphs appear in all copies
 of this software.

 IN NO EVENT SHALL THE UNIVERSITY OF MARYLAND BE LIABLE TO ANY PARTY
 FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
 THE UNIVERSITY OF MARYLAND HAS BEEN ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 THE UNIVERSITY OF MARYLAND SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 MARYLAND HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 ENHANCEMENTS, OR MODIFICATIONS.

*/


import java.lang.reflect.Array;
import ptolemy.actor.IOPort;

//////////////////////////////////////////////////////////////////////////
//// CSDFEdgeWeight
/** Information associated with an CSDF edge.
CSDFEdgeWeights are objects associated with {@link ptolemy.graph.Edge}s
that represent CSDF edges in {@link ptolemy.graph.Graph}s.
This class caches frequently-used data associated with CSDF edges.
It is also useful for intermediate CSDF graph representations
that do not correspond to Ptolemy II models, and performing graph
transformations (e.g., vectorization and retiming).
It is intended for use with analysis/synthesis algorithms that operate
on generic graph representations of CSDF models.

@author Michael Rinehart, Shuvra S. Bhattacharyya
@version $Id: CSDFEdgeWeight.java,v 1.2 2002/07/16 18:37:49 myko Exp $
@see ptolemy.graph.Edge
*/

public class CSDFEdgeWeight extends DIFEdgeWeight {

    /** Construct an edge weight for a homogeneous, zero-delay edge.
     */
    public CSDFEdgeWeight() {
        int[] singleRate = {0};
        setProductionRate(singleRate);
        setConsumptionRate(singleRate);
        setSinkPort(null);
        setSourcePort(null);
        setDelay(0);
    }

    /** Construct an edge weight for a specified token production rate,
     *  token consumption rate, and delay.
     *  @param productionRates The token production rates.
     *  @param consumptionRates The token consumption rates.
     *  @param delay The delay.
     */
    public CSDFEdgeWeight(int[] productionRates,
                          int[] consumptionRates, int delay) {
        setDelay(delay);
        setSourcePort(null);
        setSinkPort(null);
        setCSDFProductionRates(productionRates);
        setCSDFConsumptionRates(consumptionRates);
    }

    /** Construct an edge weight for a specified source port, sink port,
     *  token production rate, token consumption rate, and delay.
     *  The source port and sink port are ports (in a Ptolemy II CSDF
     *  model) that correspond to this edge.
     *  @param sourcePort The source port.
     *  @param sinkPort The sink port.
     *  @param productionRates The token production rates.
     *  @param consumptionRates The token consumption rates.
     *  @param delay The delay.
     */
    public CSDFEdgeWeight(IOPort sourcePort, IOPort sinkPort,
                          int[] productionRates,
                          int[] consumptionRates, int delay) {
        setSourcePort(sourcePort);
        setSinkPort(sinkPort);
        setDelay(delay);
        setCSDFProductionRates(productionRates);
        setCSDFConsumptionRates(consumptionRates);
    }

    ///////////////////////////////////////////////////////////////////
    ////                         public methods                    ////

    /** The sum of the token consumption rates over all phases is returned.
     *  @return The period token consumption rate.
     */
    public int consumptionPeriodRate() {
        int i, sum = 0, length = consumptionPhaseCount();

        for (i = 0; i < length; i++)
            sum += getCSDFConsumptionRate(i);

        return sum;
    }

    /** Returns the number of the phases on the consumed end of the edge
     *  @return The number of phases on the consumed end of the edge.
     */
    public int consumptionPhaseCount() {
        return ((int[])_getConsumptionRate()).length;
    }

    /** Returns the token consumption rate of the specified phase
     *  @param phase The phase.
     *  @return The token consumption rate.
     */
    public int getCSDFConsumptionRate(int phase) {
        return getCSDFConsumptionRates()[phase];
    }

    /** Get CSDF consumption rates for all phases.
     *  @return The rates in <code>int[]</code>.
     */
    public int[] getCSDFConsumptionRates() {
        return (int[])_getConsumptionRate();
    }

    /** Returns the token production rate of the specified phase
     *  @param phase The phase.
     *  @return The token production rate.
     */
    public int getCSDFProductionRate(int phase) {
        return getCSDFProductionRates()[phase];
    }

    /** Get CSDF production rates for all phases.
     *  @return The rates in <code>int[]</code>.
     */
    public int[] getCSDFProductionRates() {
        return (int[])_getProductionRate();
    }

    /** The sum of the token production rates over all phases is returned.
     *  @return The period token production rate.
     */
    public int productionPeriodRate() {
        int i, sum = 0, length = productionPhaseCount();

        for (i = 0; i < length; i++)
            sum += getCSDFProductionRate(i);

        return sum;
    }

    /** Returns the number of the phases on the produced end of the edge
     *  @return The number of phases on the produced end of the edge.
    **/
    public int productionPhaseCount() {
        return ((int[])_getProductionRate()).length;
    }

    /** Set the token consumption rates
     *  @param consumptionRates The new token consumption rates as an array.
     */
    public void setCSDFConsumptionRates(int[] consumptionRates) {
        _setConsumptionRate(consumptionRates);
    }

    /** Set the token consumption rate of the specified phase
     *  @param phase The phase to modify.
     *  @param consumptionRate The new consumption rate.
     */
    public void setCSDFConsumptionRate(int phase, int consumptionRate) {
        getCSDFConsumptionRates()[phase] = consumptionRate;
    }

    /** Set the token production rates
     *  @param productionRates The new token production rates as an array.
     */
    public void setCSDFProductionRates(int[] productionRates) {
        _setProductionRate(productionRates);
    }

    /** Set the token production rate of the specified phase
     *  @param phase The phase to modify.
     *  @param productionRate The new token production rates.
     */
    public void setCSDFProductionRate(int phase, int productionRate) {
        getCSDFProductionRates()[phase] = productionRate;
    }

    /** Return a string representation of the edge weight. This string
     *  representation is in the following form:
     *  <p>
     *  <em> [productionRate1 productionRate2 ...]
     *  [consumptionRate1 consumptionRate2 ...] delay </em>
     *  <p>
     * @return the string representation of the edge weight.
     */
    public String toString() {
        StringBuffer str = new StringBuffer("[");
        int i, clength = consumptionPhaseCount(),
            plength = productionPhaseCount();

        for (i = 0; i < plength; i++) {
            str.append(getCSDFProductionRate(i));
            if (i < (plength - 1)) str.append(" ");
        }
        str.append("]");

        for (i = 0; i < clength; i++) {
            str.append(getCSDFConsumptionRate(i));
            if (i < (clength - 1)) str.append(" ");
        }
        str.append("] ").append(getDelay());

        return str.toString();
    }
}
