/* Information associated with an CSDF graph.

 Copyright (c) 2002 The University of Maryland. All rights reserved.
 Permission is hereby granted, without written agreement and without
 license or royalty fees, to use, copy, modify, and distribute this
 software and its documentation for any purpose, provided that the above
 copyright notice and the following two paragraphs appear in all copies
 of this software.

 IN NO EVENT SHALL THE UNIVERSITY OF MARYLAND BE LIABLE TO ANY PARTY
 FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
 THE UNIVERSITY OF MARYLAND HAS BEEN ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 THE UNIVERSITY OF MARYLAND SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 MARYLAND HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 ENHANCEMENTS, OR MODIFICATIONS.

*/


import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Collection;

import ptolemy.actor.AtomicActor;
import ptolemy.actor.CompositeActor;
import ptolemy.domains.sdf.kernel.SDFDirector;
import ptolemy.domains.sdf.kernel.SDFScheduler;
import ptolemy.actor.GraphReader;
import ptolemy.kernel.util.StreamListener;
import ptolemy.math.Fraction;


//////////////////////////////////////////////////////////////////////////
//// CSDFGraph
/** Information associated with an CSDF graph.
This class caches frequently-used data associated with CSDF graphs.
It is also useful for intermediate CSDF graph representations
that do not correspond to Ptolemy II models, and performing graph
transformations (e.g. convertion to and manipulation of single-rate graphs).
It is intended for use with analysis/synthesis algorithms that operate
on generic graph representations of CSDF models.
<p>
CSDFGraph nodes and edges have weights of type {@link CSDFNodeWeight} and \
{@link CSDFEdgeWeight}, respectively.

@author Michael Rinehart, Ming-Yung Ko, Shuvra S. Bhattacharyya
@version $Id: CSDFGraph.java,v 1.4 2002/09/18 16:06:24 myko Exp $
@see CSDFEdgeWeight
@see CSDFNodeWeight
*/

public class CSDFGraph extends DIFGraph {

    /** Construct an empty CSDF graph.
     */
    public CSDFGraph() {
        super();
        _repetitions = new HashMap();
    }

    /** Construct an empty CSDF graph with enough storage allocated for the
     *  specified number of nodes.
     *  @param nodeCount The number of nodes.
     */
    public CSDFGraph(int nodeCount) {
        super(nodeCount);
        _repetitions = new HashMap();
    }

    /** Construct an empty CSDF graph with enough storage allocated for the
     *  specified number of edges, and number of nodes.
     *  @param nodeCount The integer specifying the number of nodes
     *  @param edgeCount The integer specifying the number of edges
     */
    public CSDFGraph(int nodeCount, int edgeCount) {
        super(nodeCount, edgeCount);
        _repetitions = new HashMap();
    }

    ///////////////////////////////////////////////////////////////////
    ////                         public methods                    ////

    /** Check node's repetitions. Return <code>true</code> if the repetitions
     *  have been evaluated, otherwise <code>false</code>. This method is
     *  created to cooperate
     *  new node addition.
     *  @param node The given node
     *  @returns <code>true</code> if repetitions evaluated,
     *  <code>false</code> otherwise
     */
    public boolean checkRepetitions(Node node) {
        return _repetitions.containsKey(node);
    }

    /** Compute repetitions vector for a given CSDF graph.  The repetitions
     *  will be returned as a HashMap, with {@link ptolemy.graph.Node} as
     *  the key(s) and {@link java.lang.Integer} as the repetition value(s).
     *  @param graph The CSDF graph whose repetitions will be computed
     */
    public void computeRepetitions() {
        int nodeCount = nodeCount();
        int edgeCount = edgeCount();

        if (connectedComponents().size() > 1)
            throw new RuntimeException("Graph not connected");

        _rept = new int[nodeCount];
        _denominators = new int[nodeCount];
        _rept[0] = _denominators[0] = _multiplier = 1;

        _computeNodeRepetition(node(0));
        for (int i = 0; i < nodeCount; i++ ) {
            _rept[i] *= (_multiplier / _denominators[i]);
        }
        Iterator edgeIter = edges().iterator();
        while (edgeIter.hasNext()) {
            Edge arc = (Edge)edgeIter.next();
            int srcRept = _rept[nodeLabel(arc.source())];
            int snkRept = _rept[nodeLabel(arc.sink())];
            int prod = ((CSDFEdgeWeight)arc.getWeight()).
                       productionPeriodRate();
            int cons = ((CSDFEdgeWeight)arc.getWeight()).
                       consumptionPeriodRate();
            if ((srcRept * prod) != (snkRept * cons))
                throw new RuntimeException("Inconsistent sample rates");
        }
        int rep_gcd = _rept[0];
        for (int i = 0; i < nodeCount; i++) {
            rep_gcd = Fraction.gcd(rep_gcd, _rept[i]);
        }
        if ((nodeCount > 1) && (rep_gcd > 1)) {
            throw new RuntimeException("Inconsistent sample rates");
        }
        for (int i = 0; i < nodeCount; i++) {
            _repetitions.put(node(i), new Integer(_rept[i]));
        }
    }

    /** Get the repetition count of a given CSDF node in this graph.
     *  If the repetitions value has not yet been computed, then first
     *  compute the repetitions vector for the entire graph.
     *  @param node The CSDF node.
     *  @return The repetition count.
     */
    public int getRepetitions(Node node) {
        if ((_repetitions == null) || !_repetitions.containsKey(node)) {
            computeRepetitions();
        }
        if (!_repetitions.containsKey(node)) {
            throw new RuntimeException("Could not compute repetition count "
                    + "for the specified node. Dumps of the node and graph "
                    + "follow.\n" + node.toString() + "\n" + toString());
        }
        else return ((Integer)(_repetitions.get(node))).intValue();
    }

    /** Remove a node from the CSDF graph. This method overrides
     *  <code>removeNode()</code> in <code>Graph</code>. Besides removing
     *  node from graph, it removes the entry of repetitions map.
     *  @param node The node to remove
     * @return True if the node was removed.
     */
    public boolean removeNode(Node node) {
        boolean result = super.removeNode(node);
        if (result) {
            if (_repetitions.containsKey(node)) {
                _repetitions.remove(node);
            } else {
                result = false;
            }
        }
        return result;
    }

    /** Set repetition counts for a node. Sample rate consistency will be
     *  checked. Therefore, proper sample rates should be assigned before
     *  calling this method.
     *  @param node The given node
     *  @param count The given repetition counts
     *  @exception IllegalArgumentException If the node is not contained in
     *             the graph.
     *  @exception RuntimeException Graph not connected
     *  @exception RuntimeException Inconsistent sample rates
     */
    public void setRepetitions(Node node, int count) {
        if (!containsNode(node))
            throw new IllegalArgumentException("The node is not "
                    + "contained in the graph.");

        // Special case: graph has only one node and no edges. This could
        // happen in clustering when all nodes are clustered into a single
        // root node.
        if (nodeCount() == 1) {
            _repetitions.put(node, new Integer(1));

        // Normal case. Graph is checked connected first to enable consistent
        // sample rates checking.
        } else {
            if (incidentEdges(node).isEmpty())
                throw new RuntimeException("Graph not connected");
            if (!_checkConsistency(node, count))
                throw new RuntimeException("Inconsistent sample rates");
            _repetitions.put(node, new Integer(count));
        }
    }

    /** Output a string representation the CSDF graph. In this string
     *  representation, numbers of nodes and edges are listed (on the
     *  first line), and then the edges are listed (one per line)
     *  in the format
     *  <p>
     *  <em> sourceActor sinkActor tokenProductionRate tokenConsumptionRate
     *       delay </em>
     *  <p>
     *  @return The string representation.
     */
    public String toString() {
       StringBuffer result = new StringBuffer();
       result.append(nodeCount() + " " + edgeCount() + "\n");

       Iterator edgeIter = edges().iterator();
       while (edgeIter.hasNext()) {
           result.append(((Edge)(edgeIter.next())) + "\n");
       }
       return result.toString();
    }

    /** Verify edge weight for CSDF graph.
     *  @param weight The edge weight to verify.
     *  @return True if the given edge weight is valid for CSDF graph.
     */
    public boolean validEdgeWeight(Object weight) {
        return weight instanceof CSDFEdgeWeight;
    }

    /** Verify node weight for CSDF graph.
     *  @param weight The node weight to verify.
     *  @return True if the given node weight is valid for CSDF graph.
     */
    public boolean validNodeWeight(Object weight) {
        return weight instanceof CSDFNodeWeight;
    }

    ///////////////////////////////////////////////////////////////////
    ////                       protected methods                   ////

    /** Compute the repetitions vector assuming that this CSDF graph
     *  has been converted from a Ptolemy II SDF model.
     *  The repetitions vector is returned in the form of a HashMap.
     *  The keys in the map are {@link ptolemy.graph.Node}s of the CSDFGraph,
     *  and the values are Integer representations of the associated
     *  repetitions vector values.
     *  @param compositeActor The Ptolemy II representation of the CSDF graph.
     *  @return The repetitions vector in the form of a HashMap.
     */
    protected HashMap _getRepetitionsMap(CompositeActor compositeActor) {
        HashMap result = new HashMap(nodeCount());
        SDFScheduler scheduler = null;
        try {
            SDFDirector director = new SDFDirector(compositeActor,
                    "getRepetitionsMapDirector");
            scheduler = new SDFScheduler(director,
                    "getRepetitionsMapScheduler");
        } catch (Exception exception) {
            String identifier = (compositeActor == null) ? "null" :
                    compositeActor.toString();
            throw new RuntimeException("Could not initialize scheduler for "
                    + "composite actor '" + identifier + ".'\n"
                    + exception.getMessage());
        }

        if (_debugging) {
            // Enable debugging output for the CSDFScheduler
            StreamListener schedulerListener = new StreamListener();
            scheduler.addDebugListener(schedulerListener);
        }

        Iterator nodeIter = nodes().iterator();
        while (nodeIter.hasNext()) {
            Node node = (Node)(nodeIter.next());
            int firingCount = 0;
            CSDFNodeWeight weight = (CSDFNodeWeight)(node.getWeight());
            try {
                firingCount = scheduler.getFiringCount(
                        (AtomicActor)(weight.getComputation()));
            } catch (Exception exception) {
                String identifier = (weight == null) ? "null" :
                        weight.toString();
                throw new RuntimeException("Could not compute repetitions for "
                        + "actor '" + identifier + ".'\n"
                        + exception.getMessage());
            }
            result.put(node, new Integer(firingCount));
        }
        return result;
    }

    ///////////////////////////////////////////////////////////////////
    ////                       private methods                   //////

    /** Given an node and a new repetitions, check the consistency of
     *  sampling rates.
     */
    private boolean _checkConsistency(Node node, int count) {
        boolean consistency = true;

        Iterator edges = incidentEdges(node).iterator();
        while (edges.hasNext()) {

            Edge incidentEdge = (Edge)edges.next();
            int pRate = ((CSDFEdgeWeight)incidentEdge.getWeight()).
                    productionPeriodRate();
            int cRate = ((CSDFEdgeWeight)incidentEdge.getWeight()).
                    consumptionPeriodRate();
            int production;
            int consumption;
            Node source = incidentEdge.source();
            Node sink = incidentEdge.sink();


            // compute total number of samples exchanged
            if (node == source) {
                production  = count * pRate;
                // Evluate total consumed tokens. Before doing that,
                // check repetitions of sink first. If it's unset, make
                // consumption equal to production. Sink is assumed to
                // have a repetitions in the future.
                consumption = (!checkRepetitions(sink)) ? production :
                        getRepetitions(sink) * cRate;
            } else {
                consumption = count * cRate;
                // Evluate total produced tokens. Before doing that,
                // check repetitions of source first. If it's unset, make
                // production equal to consumption. Source is assumed to
                // have a repetitions in the future.
                production = (!checkRepetitions(source)) ? consumption :
                        getRepetitions(source) * pRate;
            }
            // check consistency
            if (production != consumption) {
                consistency = false;
                break;
            }
        }
        return consistency;
    }

    /** This method computes a node's repetition by recursion.  The
     *  repetition is saved in the private static _rept[].
     */
    private void _computeNodeRepetition(Node node) {
        int prod, cons;

        Iterator outEdgeIter = outputEdges(node).iterator();
        while (outEdgeIter.hasNext()) {
            Edge nextOut = (Edge)outEdgeIter.next();
            Node sink = nextOut.sink();
            prod = ((CSDFEdgeWeight)nextOut.getWeight()).
                   productionPeriodRate();
            cons = ((CSDFEdgeWeight)nextOut.getWeight()).
                   consumptionPeriodRate();
            if (_reptEdge(node, prod, sink, cons)) {
                _computeNodeRepetition(sink);
            }
        }

        Iterator inEdgeIter = inputEdges(node).iterator();
        while (inEdgeIter.hasNext()) {
            Edge nextIn = (Edge)inEdgeIter.next();
            Node src = nextIn.source();
            prod = ((CSDFEdgeWeight)nextIn.getWeight()).
                   productionPeriodRate();
            cons = ((CSDFEdgeWeight)nextIn.getWeight()).
                   consumptionPeriodRate();
            if (_reptEdge(node, cons, src, prod)) {
                _computeNodeRepetition(src);
            }
        }
    }

    /** Check that an edge's sink node repetition is evaluated or not.
     */
    private boolean _reptEdge(Node p, int pParam, Node q, int qParam) {
        int pIndx = nodeLabel(p);
        int qIndx = nodeLabel(q);
        int gcd;
        if (_rept[qIndx] == 0) {
            _rept[qIndx] = _rept[pIndx] * pParam;
            _denominators[qIndx] = _denominators[pIndx] * qParam;
                // reduce both numbers by their GCD
            gcd = Fraction.gcd(_rept[qIndx], _denominators[qIndx]);
            _rept[qIndx] /= gcd;
            _denominators[qIndx] /= gcd;
            _multiplier = Fraction.lcm(_multiplier, _denominators[qIndx]);
            return true;
        } else {
            return false;
        }
    }

    ///////////////////////////////////////////////////////////////////
    ////                       private variables                   ////

    // Flag for local debugging output.
    private boolean _debugging = false;

    private static int[] _denominators;

    private int _multiplier;

    // The repetitions vector of the CSDF graph. The keys of this HashMap are
    // CSDF nodes (of type Node), and the values are Integers.
    private HashMap _repetitions = null;

    private static int[] _rept;
}
