/* Schedule for Node.

 Copyright (c) 2001-2002 The University of Maryland.
 All rights reserved.
 Permission is hereby granted, without written agreement and without
 license or royalty fees, to use, copy, modify, and distribute this
 software and its documentation for any purpose, provided that the above
 copyright notice and the following two paragraphs appear in all copies
 of this software.

 IN NO EVENT SHALL THE UNIVERSITY OF MARYLAND BE LIABLE TO ANY PARTY
 FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
 THE UNIVERSITY OF MARYLAND HAS BEEN ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 THE UNIVERSITY OF MARYLAND SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 MARYLAND HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 ENHANCEMENTS, OR MODIFICATIONS.

                                        PT_COPYRIGHT_VERSION_2
                                        COPYRIGHTENDKEY
*/


import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import ptolemy.actor.Actor;
import ptolemy.actor.AtomicActor;
import ptolemy.actor.CompositeActor;
import ptolemy.actor.sched.ScheduleElement;
import ptolemy.actor.sched.Schedule;
import ptolemy.actor.sched.Firing;

//////////////////////////////////////////////////////////////////////////
//// CSDFSchedule
/**
A custom schedule of Node, instead of Firing. This class encloses one Schedule
and two HashMap. The HashMap are for mapping between Firing and Node.

@author Mingyung Ko
@version $Id: CSDFSchedule.java,v 1.6 2002/09/20 16:49:54 myko Exp $
@see ptolemy.actor.sched.Schedule
@see ptolemy.actor.sched.ScheduleElement
*/

public class CSDFSchedule {

    //// Constructor
    public CSDFSchedule() {
        _iterandDelimiter = " ";
        _firingMap  = new HashMap();
        _nodeMap    = new HashMap();
        _schedule = new Schedule();
    }

    ///////////////////////////////////////////////////////////////////
    ////                         public methods                    ////

    /** Insert the specified schedule element at the specified position
     *  in the schedule list.
     *  @param index The index at which the specified element is to be
     *         inserted.
     *  @param element The schedule element to add.
     */
    public void add(int index, ScheduleElement element) {
        _schedule.add(index, element);
    }

    /** Append the specified schedule element to the end of the schedule list.
     *  @param element The schedule element to add.
     */
    public void add(ScheduleElement element) {
        _schedule.add(element);
    }

    /** Return appearance count of a given node.
     *  @param node The given node
     *  @return The appearance count
     */
    public int appearanceCount(Node node) {
        return ((List)_nodeMap.get(node)).size();
    }

    /** Sequence of <code>Firing</code>s that will be executed in order.
     *  @return The iterator over firings.
     */
    public Iterator firingIterator() {
        return _schedule.firingIterator();
    }

    /** Return all the Firing correspondent to the given node in sequence.
     *  Namely, each firing represents an appearance in looped schedules.
     *  @param node The given node.
     *  @return The firings in the form of array.
     */
    public Object[] firings(Node node) {
        return ((List)_nodeMap.get(node)).toArray();
    }

    /** Get ScheduleElement with given position.
     *  @param index The given position.
     *  @return The ScheduleElement at position "index".
     */
    public ScheduleElement get(int index) {
        return _schedule.get(index);
    }

    /** Return the node with given Firing.
     *  @param firing The given Firing.
     *  @return The correspondent node.
     */
    public Node getNode(Firing firing) {
        return (Node)(_firingMap.get(firing));
    }

    /** Return an iterator over all content ScheduleElement's.
     *  @return Iterator over content ScheduleElements.
     */
    public Iterator iterator() {
        return _schedule.iterator();
    }

    /** Get the lexical order of nodes. Lexical order is the appearing
     *  sequence of nodes in schedules.
     *  @return The lexical order of nodes
     */
    public Object[] lexicalOrder() {
        ArrayList order = new ArrayList();
        _lexicalOrder(order, _schedule);
        return order.toArray();
    }

    /** Return the maximum appearance count over all nodes.
     *  @return The maximum appearance count
     */
    public int maxAppearanceCount() {
        int biggest = 0;

        Iterator nodes = _nodeMap.keySet().iterator();
        while (nodes.hasNext()) {
            int count = ((List)_nodeMap.get(nodes.next())).size();
            if (biggest < count)
                biggest = count;
        }

        return biggest;
    }

    /** Create a <code>Firing</code> to add in the schedule. The returned
     *  Firing must be added to the schedule otherwise errors may happen.
     *  Map entries for the given node with the loop count are added to
     *  private maps.
     *  @param node The given node.
     *  @param loopCount Loop count for the node.
     *  @return The Firing created for the node.
     */
    public Firing newFiring(Node node, int loopCount) {
        Firing firing = new Firing();
        firing.setIterationCount(loopCount);

        // setup maps
        _firingMap.put(firing, node);
        if (_nodeMap.containsKey(node) == false)
            _nodeMap.put(node, new ArrayList());
        ((List)_nodeMap.get(node)).add(firing);

        return firing;
    }

    /** Create a <code>Firing</code> to add in the schedule. The node with a
     *  default loop count one will also be added to the internal map.
     *  @param node The given node.
     *  @return The Firing created for the node.
     */
    public Firing newFiring(Node node) {
        return newFiring(node, 1);
    }

    /** Return the node invocation sequence. This is so called the "linear"
     *  schedule of nodes.
     *  @return An iterator of nodes
     */
    public Iterator nodeIterator() {
        ArrayList nodes = new ArrayList();
        if (size() == 0) {
            throw new RuntimeException("Schedule not constrcuted yet");
        } else {
            Iterator firings = _schedule.firingIterator();
            while (firings.hasNext()) {
                Firing firing = (Firing)firings.next();
                Node node = getNode(firing);
                for (int i = 0; i < firing.getIterationCount(); i++) {
                    nodes.add(node);
                }
            }
        }
        return nodes.iterator();
    }

    /** Set delimiter character for schedule printing.
     *  @param delimiter The desired delimiter characters.
     *  @return The delimiter in String.
     */
    public String setIterandDelimiter(String delimiter) {
        _iterandDelimiter = delimiter;
        return _iterandDelimiter;
    }

    /** Set vectorizaion factor for the whole schedule.
     *  @param count The desired factor.
     */
    public void setVectorizationFactor(int count) {
        _schedule.setIterationCount(count);
    }

    /** Return the number of elements in this list.
     *  @return The number of elements in this list.
     */
    public int size() {
        return _schedule.size();
    }

    /** Return the standard Ptolemy II schedule associated with
     *  this CSDF schedule. This is only valid when the associated
     *  graph is derived from a Ptolemy II model.
     *  @return The standard Ptolemy II schedule.
     */
    public Schedule toSchedule() {
        _makePtolemySchedule(_schedule);
        return _schedule;
    }

    /** A <code>String</code> representation of the schedule.
     *  @return A <code>String</code> description of the schedule.
     */
    public String toString() {
        String printing = new String();
        printing += _printSchedule(_schedule);
        return printing;
    }

    /** Get vectorizaion factor for the whole schedule.
     *  @return The vectorizaion factor.
     */
    public int vectorizationFactor() {
        return _schedule.getIterationCount();
    }

    ///////////////////////////////////////////////////////////////////
    ////                       protected methods                   ////

    /** The main routine of toString(). This method recursively prints
     *  a schedule. The schedule is traversed top down to expand the
     *  schedule hierarchy.
     *  @param iterand The schedule itreand.
     *  @return The schedule enclosing the argument iterand.
     */
    protected String _printSchedule(ScheduleElement iterand) {
        String printing = new String();

        int iteration = iterand.getIterationCount();

        // Print only counts greater than 1. An opening parenthesis and a
        // space is needed to follow the count.
        if (iteration > 1) {
            printing += "(" + iteration + " ";
        }

        // "iterand" is a Firing.
        if (iterand.getClass() == (new Firing()).getClass()) {
            // Node is the actual iterand.
            Node realIterand = getNode((Firing)iterand);
            CSDFNodeWeight nodeWeight = (CSDFNodeWeight)realIterand.getWeight();
            printing += nodeWeight.getComputation().toString();

        // "iterand" is a Schedule.
        } else {
            Iterator elements = ((Schedule)iterand).iterator();
            while (elements.hasNext()) {
                ScheduleElement nextElement = (ScheduleElement)elements.next();
                printing += _printSchedule((ScheduleElement)nextElement);
                if (elements.hasNext()) {
                    printing += _iterandDelimiter;
                }
            }
        }

        // Print the closing parenthesis.
        if (iteration > 1) {
            printing += ")";
        }

        return printing;
    }

    ///////////////////////////////////////////////////////////////////
    ////                      private methods                     ////

    // main routine for getLexicalOrder()
    private void _lexicalOrder(ArrayList order, ScheduleElement element) {
        if (element instanceof Firing) {
            order.add(getNode((Firing)element));
        } else {
            Iterator elements = ((Schedule)element).iterator();
            while (elements.hasNext()) {
                _lexicalOrder(order, (ScheduleElement)elements.next());
            }
        }
    }

    /* Make a valid Ptolemy schedule from _schedule. If a Firing of a node
     * does not have a corresponding Actor, then an Actor will be created
     * for the Firing and serve as the node's computation.
     */
    private void _makePtolemySchedule(ScheduleElement element) {
        // If the element is a Firing
        if (element instanceof Firing) {
            Object nComp = ((CSDFNodeWeight)getNode((Firing)element).getWeight()).
                    getComputation();
            // If the node's computation is an AtomicActor, copy it to the
            // new Firing. Otherwise, new an Actor.
            if (nComp instanceof AtomicActor)  {
                ((Firing)element).setActor((AtomicActor)nComp);
            } else if (nComp instanceof CompositeActor) {
                ((Firing)element).setActor((CompositeActor)nComp);
            } else {
                throw new RuntimeException("Node does not have Actor" +
                        " computation.");
            }

        // If the element is a Schedule.
        } else {
            Iterator subElements = ((Schedule)element).iterator();
            while (subElements.hasNext()) {
                ScheduleElement subElement =
                        (ScheduleElement)subElements.next();
                _makePtolemySchedule(subElement);
            }
        }
    }

    ///////////////////////////////////////////////////////////////////
    ////                       private variables                   ////

    // the actual ptolemy schedule
    private Schedule _schedule;

    // A map from Firing to Node
    private HashMap _firingMap;

    // A map from Node to ArrayList of Firing
    private HashMap _nodeMap;

    // Iterand delimiter for schedule printing
    private String _iterandDelimiter;

}

