/* A base CSDF scheduler

 Copyright (c) 2001-2002 The University of Maryland. All rights reserved.
 Permission is hereby granted, without written agreement and without
 license or royalty fees, to use, copy, modify, and distribute this
 software and its documentation for any purpose, provided that the above
 copyright notice and the following two paragraphs appear in all copies
 of this software.

 IN NO EVENT SHALL THE UNIVERSITY OF MARYLAND BE LIABLE TO ANY PARTY
 FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
 THE UNIVERSITY OF MARYLAND HAS BEEN ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 THE UNIVERSITY OF MARYLAND SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 MARYLAND HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 ENHANCEMENTS, OR MODIFICATIONS.

*/


import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import ptolemy.actor.sched.Firing;

//////////////////////////////////////////////////////////////////////////
//// CSDFScheduler
/** A base class for CSDF schedulers. This class extends
{@link ptolemy.graph.analysis.Analysis}. The advantage of this inheritance
is that schedule re-computation is necessary on if the graph is obsolete.

@author Mingyung Ko
@version $Id: CSDFScheduler.java,v 1.6 2002/07/19 16:56:45 myko Exp $
*/

public abstract class CSDFScheduler extends Analysis {

    /** Constructor for a given graph.
     *
     *  @param graph The given CSDF graph.
     */
    public CSDFScheduler(CSDFGraph graph) {
        super(graph);
    }

    ///////////////////////////////////////////////////////////////////
    ////                         public methods                    ////

    /** Test compatibility in terms of the class of graph. Only
     *  {@link CSDFGraph}s is compatible with <code>CSDFScheduler</code>.
     *
     *  @param graph The given graph to test compatibility.
     *  @return <code>True</code> if compatible; <code>false</code> otherwise.
     */
    public boolean compatible(Graph graph) {
        return (graph instanceof CSDFGraph);
    }

    /** Configuration of the scheduler.
     *
     *  @return A text describing the configuration.
     */
    public String configuration() {
        String configuration = new String();
        configuration += "Configuration for the scheduler: " +
                this.getClass().getName();

        configuration += "\nSingle Appearance Schedule: ";
        if (schedule().maxAppearanceCount() == 1) {
            configuration += "Yes";
        } else {
            configuration += "No";
        }

        return configuration;
    }

    /** Return a incompatibility description in terms of the
     *  class of graph.
     *
     *  @param graph The given graph.
     *  @return A description of incompatibility.
     */
    public String incompatibilityDescription(Graph graph) {
        String result = new String();
        if (!(graph instanceof CSDFGraph)) {
            result = "The given graph of class " + graph.getClass().getName() +
                     " is not an instance of CSDFGraph";
        }
        return result;
    }

    /** Return the schedule in the form of {@link CSDFSchedule}.
     *
     *  @return The result schedule.
     */
    public final CSDFSchedule schedule() {
        return (CSDFSchedule)result();
    }

    /** A desrciption of the scheduler.
     *
     *  @return A text description.
     */
    public String toString() {
        return new String("A base CSDFGraph scheduler.");
    }


}
