/* A base scheduler using clustering strategies.

Copyright (c) 2002 The University of Maryland.

Permission is hereby granted, without written agreement and without
license or royalty fees, to use, copy, modify, and distribute this
software and its documentation for any purpose, provided that the above
copyright notice and the following two paragraphs appear in all copies
of this software.

IN NO EVENT SHALL THE UNIVERSITY OF MARYLAND BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
THE UNIVERSITY OF MARYLAND HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

THE UNIVERSITY OF MARYLAND SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
MARYLAND HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS.

@AcceptedRating Red

*/


import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;

import ptolemy.actor.sched.Firing;
import ptolemy.actor.sched.Schedule;
import ptolemy.actor.sched.ScheduleElement;

//////////////////////////////////////////////////////////////////////////
//// ClusterScheduler
/** A base scheduler using clustering strategies.
<p>
Cluster operations like
{@link SDFGraphs#clusterNodes(SDFGraph, Collection, Node)}
change the original graph structure. The changing behavior is not
desired and graph cloning is necessary. All clustering actions are
therefore performed on the cloned graph and not on the original one
anymore. By this way, the original graph is left untouched and
the topology is kept.

@author Mingyung Ko
@version $Id: ClusterScheduler.java,v 1.5 2002/07/23 16:57:10 myko Exp $
*/

abstract public class ClusterScheduler extends SDFScheduler {

    /** Constructor for a given SDF graph.
     *
     *  @param graph The given SDF graph.
     */
    public ClusterScheduler(SDFGraph graph) {
        super(graph);
        // Clone the original/flattened graph.
        _clusteredGraph = (SDFGraph)graph.clone();
        // Initialize the map for maintaining super nodes.
        _superNodesMap = new HashMap();
        Iterator nodes = _clusteredGraph.nodes().iterator();
        while (nodes.hasNext())
            _superNodesMap.put(nodes.next(), null);
    }

    ///////////////////////////////////////////////////////////////////
    ////                       protected methods                   ////

    /** Get the cluster graph. Clustered graph is different from the
     *  original graph. It is the graph on which various cluster
     *  operations are working. The original graph is kept and left
     *  untouched during clustering.
     *
     *  @return The clustered graph.
     */
    final protected SDFGraph _getClusteredGraph() {
        return _clusteredGraph;
    }

    /** Get the subgraph associated with the super node.
     *
     *  @param superNode The super node.
     *  @return The associated subgraph in {@link SDFGraph}.
     */
    final protected SDFGraph _getSubgraph(Node superNode) {
        return (SDFGraph)_superNodesMap.get(superNode);
    }

    /** Test the given node super or not.
     *
     *  @param node The given node.
     *  @return True if it is a super node.
     */
    final protected boolean _isSuperNode(Node node) {
        return (_superNodesMap.get(node) != null);
    }

    /** Cluster a collection of nodes into a super node.
     *
     *  @param graph The graph whose nodes are to cluster.
     *  @param nodeCollection The collection of nodes to cluster.
     *  @param superNode The new super node to add in graph.
     */
    final protected void _clusterNodes(
            SDFGraph graph, Collection nodeCollection, Node superNode) {
        SDFGraph subgraph =
                SDFGraphs.clusterNodes(graph, nodeCollection, superNode);
        _setSuperNode(superNode, subgraph);
    }

    ///////////////////////////////////////////////////////////////////
    ////                        private methods                    ////

    /*  Set a node as super node with the associated graph.
     *
     *  @param superNode The node to set as super.
     *  @param graph The associated subgraph in {@link SDFGraph}.
     */
    private void _setSuperNode(Node superNode, SDFGraph subgraph) {
        if (_superNodesMap.containsKey(superNode))
            throw new RuntimeException("The super node already exists.");
        else
            _superNodesMap.put(superNode, subgraph);
    }

    ///////////////////////////////////////////////////////////////////
    ////                       private variables                   ////

    /* A map from super node to the induced subgraph. */
    private HashMap _superNodesMap;

    /* The cloned graph on which clustering is performed */
    private SDFGraph _clusteredGraph;
}
