/* Information associated with an DIF edge.

 Copyright (c) 2002 The University of Maryland
 All rights reserved.
 Permission is hereby granted, without written agreement and without
 license or royalty fees, to use, copy, modify, and distribute this
 software and its documentation for any purpose, provided that the above
 copyright notice and the following two paragraphs appear in all copies
 of this software.

 IN NO EVENT SHALL THE UNIVERSITY OF MARYLAND BE LIABLE TO ANY PARTY
 FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
 THE UNIVERSITY OF MARYLAND HAS BEEN ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 THE UNIVERSITY OF MARYLAND SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 MARYLAND HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 ENHANCEMENTS, OR MODIFICATIONS.

*/


import ptolemy.actor.IOPort;

//////////////////////////////////////////////////////////////////////////
//// DIFEdgeWeight
/** Information associated with an DIF edge.
DIFEdgeWeights are objects associated with {@link ptolemy.graph.Edge}s
that represent DIF edges in {@link ptolemy.graph.Graph}s.
This class caches frequently-used data associated with DIF edges.
It is also useful for intermediate DIF graph representations
that do not correspond to Ptolemy II models, and performing graph
transformations (e.g., vectorization and re-timing).
It is intended for use with analysis/synthesis algorithms that operate
on generic graph representations of DIF models.

@author Michael Rinehart, Shuvra S. Bhattacharyya
@version $Id: DIFEdgeWeight.java,v 1.3 2002/09/22 22:17:41 keceli Exp $
@see ptolemy.graph.Edge
*/

public class DIFEdgeWeight {

    /** Construct an edge weight for a homogeneous, zero-delay edge.
     */
    public DIFEdgeWeight() {
        _delay = 0;
        _sourcePort = null;
        _sinkPort = null;
    }

    /** Construct an edge weight for a specified token production rate,
     *  token consumption rate, and delay.
     *  @param productionRate The token production rates.
     *  @param consumptionRate The token consumption rates.
     *  @param delay The delay.
     */
    public DIFEdgeWeight(Object productionRate, Object consumptionRate,
                         int delay) {
        _productionRate = productionRate;
        _consumptionRate = consumptionRate;
        _delay = delay;
        _sourcePort = null;
        _sinkPort = null;
    }

    /** Construct an edge weight for a specified source port, sink port,
     *  token production rate, token consumption rate, and delay.
     *  The source port and sink port are ports (in a Ptolemy II DIF
     *  model) that correspond to this edge.
     *  @param sourcePort The source port.
     *  @param sinkPort The sink port.
     *  @param productionRate The token production rates.
     *  @param consumptionRate The token consumption rates.
     *  @param delay The delay.
     */
    public DIFEdgeWeight(IOPort sourcePort, IOPort sinkPort,
                         Object productionRate,
                         Object consumptionRate, int delay) {
        _productionRate = productionRate;
        _consumptionRate = consumptionRate;
        _delay = delay;
        _sourcePort = sourcePort;
        _sinkPort = sinkPort;
    }

    ///////////////////////////////////////////////////////////////////
    ////                         public methods                    ////

    /** Return the token consumption rate of the associated DIF edge.
     *  The sum of the token consumption rates over all phases is returned.
     *  @return The token consumption rate.
     */
    public Object getConsumptionRate() {
        return _consumptionRate;
    }

    /** Return the delay of the associated DIF edge.
     *  @return The delay.
     */
    public int getDelay() {
        return _delay;
    }

    /** Return the token production rate of the associated DIF edge.
     *  The sum of the token production rates over all phases is returned.
     *  @return The token production rate.
     */
    public Object getProductionRate() {
        return _productionRate;
    }

    /** Get the sink port of the associated DIF edge .
     *  @return The sink port.
     */
    public IOPort getSinkPort() {
        return _sinkPort;
    }

    /** Get the source port of the associated DIF edge .
     *  @return The source port.
     */
    public IOPort getSourcePort() {
        return _sourcePort;
    }

    /** Set the token consumption rate of the associated DIF edge.
     *  @param consumptionRate The new token consumption rates.
     */
    public void setConsumptionRate(Object consumptionRate) {
        _consumptionRate = consumptionRate;
    }

    /** Set the delay of the associated DIF edge.
     *  @param delay The new delay value.
     */
    public void setDelay(int delay) {
        _delay = delay;
    }

    /** Set the token production rate of the associated DIF edge.
     *  @param productionRate The new token production rates.
     */
    public void setProductionRate(Object productionRate) {
        _productionRate = productionRate;
    }

    /** Set the sink port of the associated DIF edge to a port (in a Ptolemy II
     *   model) that is represented by the sink of the edge.
     *  @param sinkPort The sink port.
     */
    public void setSinkPort(IOPort sinkPort) {
        _sinkPort = sinkPort;
    }

    /** Set the source port of the associated DIF edge to a port (in a Ptolemy
     *  II model) that is represented by the source of the edge.
     *  @param sourcePort The source port.
     */
    public void setSourcePort(IOPort sourcePort) {
        _sourcePort = sourcePort;
    }

    /** Return a string representation of the edge weight.
     * @return The string representation of the edge weight.
     */
    public String toString() {
        return new String("");
    }

    ///////////////////////////////////////////////////////////////////
    ////                       protected methods                   ////
    /** Return the token consumption rate of the associated DIF edge.
     *  The sum of the token consumption rates over all phases is returned.
     *  @return The token consumption rate.
     */
    protected final Object _getConsumptionRate() {
        return _consumptionRate;
    }

    /** Return the token production rate of the associated DIF edge.
     *  The sum of the token production rates over all phases is returned.
     *  @return The token production rate.
     */
    protected final Object _getProductionRate() {
        return _productionRate;
    }

    /** Set the token consumption rate of the associated DIF edge.
     *  @param consumptionRate The new token consumption rates.
     */
    protected final void _setConsumptionRate(Object consumptionRate) {
        _consumptionRate = consumptionRate;
    }

    /** Set the token production rate of the associated DIF edge.
     *  @param consumptionRate The new token production rates.
     */
    protected final void _setProductionRate(Object productionRate) {
        _productionRate = productionRate;
    }


    ///////////////////////////////////////////////////////////////////
    ////                       private variables                   ////

    // The delay of the associated DIF edge.
    private int _delay;

    // The sink IO port associated with an DIF edge if there is one
    // (otherwise, null).
    private IOPort _sinkPort;

    // The source IO port associated with an DIF edge if there is one
    // (otherwise, null).
    private IOPort _sourcePort;

    // The token consumption rate of the associated DIF edge.
    private Object _consumptionRate;

    // The token production rate of the associated DIF edge.
    private Object _productionRate;
}
