/* Information associated with an DIF graph.

 Copyright (c) 2002 The University of Maryland. All rights reserved.
 Permission is hereby granted, without written agreement and without
 license or royalty fees, to use, copy, modify, and distribute this
 software and its documentation for any purpose, provided that the above
 copyright notice and the following two paragraphs appear in all copies
 of this software.

 IN NO EVENT SHALL THE UNIVERSITY OF MARYLAND BE LIABLE TO ANY PARTY
 FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
 THE UNIVERSITY OF MARYLAND HAS BEEN ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 THE UNIVERSITY OF MARYLAND SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 MARYLAND HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 ENHANCEMENTS, OR MODIFICATIONS.

*/


import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import ptolemy.data.ArrayToken;
import ptolemy.data.DoubleMatrixToken;
import ptolemy.data.DoubleToken;
import ptolemy.data.StringToken;
import ptolemy.data.Token;
import ptolemy.data.ObjectToken;
import ptolemy.data.expr.Parameter;
import ptolemy.kernel.util.IllegalActionException;
import ptolemy.kernel.util.NameDuplicationException;
import ptolemy.kernel.util.NamedObj;


//////////////////////////////////////////////////////////////////////////
//// DIFGraph
/** Information associated with an DIF graph.
This class caches frequently-used data associated with DIF graphs.
It is also useful for intermediate DIF graph representations
that do not correspond to Ptolemy II models, and performing graph
transformations (e.g. conversion to and manipulation of single-rate graphs).
It is intended for use with analysis/synthesis algorithms that operate
on generic graph representations of CSDF models.
<p>

@author Fuat Keceli, Michael Rinehart, Shuvra S. Bhattacharyya
@version $Id: DIFGraph.java,v 1.19 2002/09/20 02:42:23 keceli Exp $
@see DIFEdgeWeight
@see DIFNodeWeight
*/

public class DIFGraph extends DirectedGraph {

    /** Construct an empty DIF graph.
     */
    public DIFGraph() {
        super();
        _containers = new HashMap();
        _setAttributeContainer(this, new NamedObj());
    }

    /** Construct an empty DIF graph with enough storage allocated for the
     *  specified number of nodes.
     *  @param nodeCount The number of nodes.
     */
    public DIFGraph(int nodeCount) {
        super(nodeCount);
        _containers = new HashMap();
        _setAttributeContainer(this, new NamedObj());
    }

    /** Construct an empty DIF graph with enough storage allocated for the
     *  specified number of edges, and number of nodes.
     *  @param nodeCount The integer specifying the number of nodes
     *  @param edgeCount The integer specifying the number of edges
     */
    public DIFGraph(int nodeCount, int edgeCount) {
        super(nodeCount, edgeCount);
        _containers = new HashMap();
        _setAttributeContainer(this, new NamedObj());
    }

    ///////////////////////////////////////////////////////////////////
    ////                       public methods                      ////

    /** Defined by {@link ptolemy.graph.Graph#addGraph}. Copies node and edge
     *  attributes of the  subgraph into this graph as well as the
     *  topology.
     *  @param graph The graph to add.
     *  @return True if this graph changed as a result of this call.
     */
    public boolean addGraph(Graph graph) {
        boolean returnValue = super.addGraph(graph);
        if(graph instanceof DIFGraph) {
            for(Iterator nodes = graph.nodes().iterator(); nodes.hasNext(); ) {
                Object node = nodes.next();
                try {
                    Object newContainer = ((NamedObj)((DIFGraph)graph)
                            ._containers.get(node)).clone();
                    _containers.put(node, newContainer);
                } catch(CloneNotSupportedException e) {
                    throw new RuntimeException("This exception should not"
                            + "  happen.");
                }
            }
            for(Iterator edges = graph.edges().iterator(); edges.hasNext(); ) {
                Object edge = edges.next();
                try {
                    Object newContainer = ((NamedObj)((DIFGraph)graph)
                            ._containers.get(edge)).clone();
                    _containers.put(edge, newContainer);
                } catch(CloneNotSupportedException e) {
                    throw new RuntimeException("This exception should not"
                            + " happen.");
                }
            }
        }
        return returnValue;       
    }

    /** Returns a clone of this graph in the form of the argument graph type
     *  (i.e., the run-time type of the returned graph is that of the
     *  argument graph). The clone has the
     *  same set of nodes and edges. Changes to the node or edge weights
     *  affect the
     *  clone simultaneously. If the run-time type of the argument graph is
     *  equal to that of this graph, then the clone is equal to this
     *  graph, as determined by {@link #equals(Object)}. However,
     *  modifications to the graph topology
     *  make the clone not equal to this graph. The same property is valid for
     *  the attributes.
     *
     *  @param graph The graph that gives the run-time type of the clone.
     *  @return A clone of this graph.
     */
    public Graph cloneAs(Graph graph) {
        Graph clone = null;
        try {
            clone = (Graph)graph.getClass().newInstance();
            clone.addGraph(this);
            if(clone instanceof DIFGraph) {
                ((DIFGraph)clone)._containers.put("graph",
                        ((NamedObj)_containers.get("graph")).clone());
            }
        } catch(InstantiationException e) {
            throw new RuntimeException("Could not create an empty graph from "
                    + " this one.\n" + e + "\n"
                    + GraphException.graphDump(this));
        } catch(IllegalAccessException e) {
            throw new RuntimeException("Could not create an empty graph from "
                    + " this one.\n" + e + "\n"
                    + GraphException.graphDump(this));
        } catch(CloneNotSupportedException e) {
            throw new RuntimeException("This exception should not happen.");
        }
        return clone;
    }

    /** Test if a graph is equal to this one. It is equal
     *  if it is of the same class, and has the same sets of nodes/edges and
     *  all the attributes are the same. For two attributes to be the same
     *  they should have the same names and same values but not necessarily the
     *  same objects.
     *
     *  @param graph The graph with which to compare this graph.
     *  @return True if the graph is equal to this one.
     */
    public boolean equals(Object graph) {
        if(super.equals(graph)) {
            try {
                if(!getAttributeDescriptions(this).equals(((DIFGraph)graph)
                           .getAttributeDescriptions(graph))) {
                    return false;
                }
                for(Iterator edges = edges().iterator(); edges.hasNext(); ) {
                    Object edge = edges.next();
                    if(!getAttributeDescriptions(edge).equals(((DIFGraph)graph)
                               .getAttributeDescriptions(edge))) {
                        return false;
                    }
                }
                for(Iterator nodes = nodes().iterator(); nodes.hasNext(); ) {
                    Object node = nodes.next();
                    if(!getAttributeDescriptions(node).equals(((DIFGraph)graph)
                               .getAttributeDescriptions(node))) {
                        return false;
                    }
                }
                return true;
            } catch(DIFException e) {
                e.printStackTrace();
                throw
                    new RuntimeException("This exception should not happen.");
            }
        }
        return false;
    }

    /** Returns the value of the attribute associated with this object in the
     *  form of a <code>ptolemy.data.Token</code>.
     *  @param object An object with runtime type of edge/node or which is
     *  this graph itself (e.g. <code>graph.getAttribute(graph)</code>).
     *  @return An object.
     *  @exception DIFException If the runtime class of this object
     *  is not an edge/node, it is not contained in the graph or it is not the
     *  graph itself.
     */
    public Object getAttribute(Object object, String name)
            throws DIFException {
        NamedObj container = _getAttributeContainer(object);
        Token returnValue = null;
        Parameter parameter = (Parameter)container.getAttribute(name);
        if(parameter != null) {
            try {
                returnValue = parameter.getToken();
            } catch(IllegalActionException e) {
                throw
                    new RuntimeException("This exception should not happen.");
            }
        } else {
            throw new DIFException("Attribute doesn't exist: " + name
                    + " in graph " + getName());
        }
        return _convertToken2Object(returnValue);
    }

    /** Returns a list of attribute names that can be found associated with
     *  this object. The list will be sorted.
     *  @param object An object with runtime type of edge/node or which is
     *  this graph itself (e.g. <code>graph.getAttribute(graph)</code>).
     *  @return A list of Strings.
     *  @exception DIFException If the runtime class of this object
     *  is not an edge/node, it is not contained in the graph or it is not the
     *  graph itself.
     */
    public List getAttributeNames(Object object) throws DIFException {
        NamedObj container = _getAttributeContainer(object);
        LinkedList names = new LinkedList();
        for(Iterator attributes = container.attributeList().iterator();
            attributes.hasNext(); ) {
            Parameter attribute = (Parameter) attributes.next();
            names.add(attribute.getName());
        }
        String[] nameArray = (String[])names.toArray(new String[0]);
        Arrays.sort(nameArray);
        return Arrays.asList(nameArray);
    }

    /** Returns the description of a graph element (node/edge) or the graph
     *  itself. This description contains the name of the object as defined by
     *  {@link #setName} method, names and values of the attributes. Attributes
     *  are sorted according to their names so this description is canonical.
     *  For this method to work properly types of the attribute values should
     *  be one of four types defined
     *  in {@link synthesis.dif.language.Writer#objectToString}. If any of the
     *  attribute values can't be read it will be skipped.
     *  @param object A graph element (node/edge) or the graph itself.
     *  @return Attribute description of the object.
     */
    public String getAttributeDescriptions(Object object) throws DIFException {
        StringBuffer buffer = new StringBuffer();
        buffer.append("Name: " + getName(object) + "\n");
        for(Iterator attributeNames = getAttributeNames(object).iterator();
            attributeNames.hasNext(); ) {
            String attributeName = (String)attributeNames.next();
            Object attribute = getAttribute(object, attributeName);
            buffer.append("Attribute name: " + attributeName + "\n"
                        + "Value: " + attribute + "\n");
        }
        return buffer.toString();
    }

    /** Returns the name of this edge/node or graph which is assigned by the
     *  {@link #setName} method. If no name or null was assigned as the name
     *  it is set to empty string.
     *  @param object An object with runtime type of edge/node or which is
     *  this graph itself (e.g. <code>graph.getName(graph)</code>).
     *  @return The name of this object.
     *  @exception DIFException If the runtime class of this object
     *  is not an edge/node, it is not contained in the graph or it is not the
     *  graph itself.
     */
    public String getName(Object object) throws DIFException {
        return _getAttributeContainer(object).getName();
    }

    /** Returns the name of this graph which is assigned by the
     *  {@link #setName} method. If no name or null was assigned as the name
     *  it is set to empty string.
     *  @return The name of this graph.
     */
    public String getName() {
        String returnValue = null;
        try {
            returnValue = _getAttributeContainer(this).getName();
        } catch(DIFException e) {
            throw new RuntimeException("This exception shouldn't happen.");
        }
        return returnValue;
    }

    /** Returns the first object that has this name specified by getName
     *  method. Returns null if no such object is found.
     *  @param name A string.
     *  @return A node/edge in this graph or this graph itself. Null if name
     *  is not found.
     */
    public Object getObject(String name) {
        Object returnValue = null;
        if(getName().equals(name)) {
            return this;
        }
        try {
            for(Iterator edges = edges().iterator(); edges.hasNext(); ) {
                Object edge = edges.next();
                if(getName(edge).equals(name)) {
                    return edge;
                }
            }
            for(Iterator nodes = nodes().iterator(); nodes.hasNext(); ) {
                Object node = nodes.next();
                if(getName(node).equals(name)) {
                    return node;
                }
            }
        } catch(DIFException e) {
            // Nothing to do...
        }
        return null;
    }

    /** Returns the hash code for this graph.
     *
     *  @return The hash code for this graph.
     */
    public int hashCode() {
        int code = super.hashCode();
        try {
            for(Iterator edges = edges().iterator(); edges.hasNext(); ) {
                Object edge = edges.next();
                code += getAttributeDescriptions(edge).hashCode();
            }
            for(Iterator nodes = nodes().iterator(); nodes.hasNext(); ) {
                Object node = nodes.next();
                code += getAttributeDescriptions(node).hashCode();
            }
            code += getAttributeDescriptions(this).hashCode();
        } catch(DIFException e) {
            throw new RuntimeException("This exception should not happen.");
        }
        return code;
    }

    /** Removes an attribute from the attribute list of an object.
     *  @param object An object with runtime type of edge/node or which is this
     *  graph (e.g. <code>graph.setName(graph, "name")</code>).
     *  @param name The name of the attribute to be deleted.
     *  @return True if there existed an attribute with the given name.
     *  @exception DIFException If the runtime class of this object
     *  is not an edge/node, it is not contained in the graph or it is not the
     *  graph itself.
     */
    public boolean removeAttribute(Object object, String name)
            throws DIFException{
        try {
            Parameter parameter = (Parameter)_getAttributeContainer(object)
                    .getAttribute(name);
            if(parameter == null) {
                return false;
            }
            parameter.setContainer(null);
        } catch(IllegalActionException e) {
            throw new RuntimeException("This exception shouldn't happen.");
        } catch(NameDuplicationException e) {
            throw new RuntimeException("This exception shouldn't happen.");
        }
        return true;
    }

    /** Remove an edge from this graph.
     * An edge that is removed from a graph can be re-inserted
     * into the graph at a later time (using {@link #addEdge(Edge)}),
     * provided that the incident nodes are still in the graph. However
     * previous attributes of the edge won't be recovered.
     * @param edge The edge to be removed.
     * @return True if the edge was removed.
     * @exception IllegalArgumentException If the edge is not contained
     * in the graph.
     */
    public boolean removeEdge(Edge edge) {
        boolean returnValue = super.removeEdge(edge);
        _containers.remove(edge);
        return returnValue;
    }

    /** Remove a node from this graph.
     * All edges incident to the node are also removed.
     * @param node The node to be removed.
     * @return True if the node was removed.
     * @exception IllegalArgumentException If the node is not contained
     * in the graph.
     */
    public boolean removeNode(Node node) {
        boolean returnValue = super.removeNode(node);
        _containers.remove(node);
        return returnValue;
    }

    /** Sets an attribute of an object. If there exist another attribute with
     *  the same name previous value is overridden.
     *  @param object An object with runtime type of edge/node or which is this
     *  graph (e.g. <code>graph.setName(graph, "name")</code>).
     *  @param name The name of the attribute to be set.
     *  @param value An object representing the value of the attribute.
     *  @return True if the attribute was not previously defined for this
     *  object.
     *  @exception DIFException If the runtime class of this object
     *  is not an edge/node, it is not contained in the graph or it is not the
     *  graph itself.
     *  @exception DIFException If a list is zero length or an array is null.
     */
    public boolean setAttribute(Object object, String name, Object value)
            throws DIFException {
        NamedObj container = _getAttributeContainer(object);
        Token token = _convertObject2Token(value);
        try {
            new Parameter(container, name, token);
        } catch(IllegalActionException e) {
            throw new RuntimeException("This exception shouldn't happen.");
        } catch(NameDuplicationException e) {
            removeAttribute(object, name);
            try {
                new Parameter(container, name, token);
            } catch(IllegalActionException e2) {
                throw new RuntimeException("This exception shouldn't happen.");
            } catch(NameDuplicationException e2) {
                throw new RuntimeException("This exception shouldn't happen.");
            }
            return false;
        }
        return true;
    }

    /** Sets the name of this edge/node or graph. If it is not set or set to
     *  null name is empty string.
     *  @param object An object with runtime type of edge/node or which is this
     *  graph (e.g. <code>graph.setName(graph, "name")</code>).
     *  @param name Name string which shouldn't include a period.
     *  @exception IllegalActionException If the object is not a node/edge
     *  in the graph or the graph itself or if the name includes a period.
     *  @exception DIFException If the runtime class of this object
     *  is not an edge/node, it is not contained in the graph or it is not the
     *  graph itself.
     */
    public void setName(Object object, String name)
            throws DIFException, IllegalActionException {
        try {
            _getAttributeContainer(object).setName(name);
        } catch(NameDuplicationException e) {
            // This exception should not be thrown by definition of NamedObj.
            throw new RuntimeException("Internal DIF compiler error:"
                    + "NameDuplicationException."
                    + "Original exception is: " + e.getMessage());
        }
    }

    /** Overrides the method in <code>ptolemy.graph.Graph</code> to copy the
     *  attributes as well. For each element copied, attributes will be copied
     *  together.
     *  @param nodes The collection of nodes; each element is a {@link Node}.
     *  @return The induced subgraph.
     */
    public Graph subgraph(Collection nodes) {
        DIFGraph subgraph = (DIFGraph)super.subgraph(nodes);
        try {
            for(Iterator subNodes = subgraph.nodes().iterator();
                subNodes.hasNext(); ) {
                Object node = subNodes.next();
                subgraph.setName(node, getName(node));
                for(Iterator attributeNames =
                        getAttributeNames(node).iterator();
                    attributeNames.hasNext(); ) {
                    String attributeName = (String)attributeNames.next();
                    subgraph.setAttribute(node, attributeName,
                            getAttribute(node, attributeName));
                }
            }
            for(Iterator edges = subgraph.edges().iterator();
                edges.hasNext(); ) {
                Object edge = edges.next();
                subgraph.setName(edge, getName(edge));
                for(Iterator attributeNames =
                        getAttributeNames(edge).iterator();
                    attributeNames.hasNext(); ) {
                    String attributeName = (String)attributeNames.next();
                    subgraph.setAttribute(edge, attributeName,
                            getAttribute(edge, attributeName));
                }
            }
        } catch (DIFException e) {
            throw new RuntimeException(e.getMessage());
        } catch (IllegalActionException e) {
            throw new RuntimeException(e.getMessage());
        }
        return subgraph;
    }

    /** Overrides the method in <code>ptolemy.graph.Graph</code> to copy the
     *  attributes as well. For each element copied, attributes will be copied
     *  together.
     *  @param nodes The subset of nodes; each element is an instance
     *  of {@link Node}.
     *  @param edges The subset of edges. Each element is an instance
     *  of {@link Edge}.
     *  @return The subgraph.
     */
    public Graph subgraph(Collection nodeCollection,
            Collection edgeCollection) {
        DIFGraph subgraph = (DIFGraph)super.subgraph(nodeCollection,
                edgeCollection);
        try {
            for(Iterator subNodes = subgraph.nodes().iterator();
                subNodes.hasNext(); ) {
                Object node = subNodes.next();
                subgraph.setName(node, getName(node));
                for(Iterator attributeNames =
                        getAttributeNames(node).iterator();
                    attributeNames.hasNext(); ) {
                    String attributeName = (String)attributeNames.next();
                    subgraph.setAttribute(node, attributeName,
                            getAttribute(node, attributeName));
                }
            }
            for(Iterator edges = subgraph.edges().iterator();
                edges.hasNext(); ) {
                Object edge = edges.next();
                subgraph.setName(edge, getName(edge));
                for(Iterator attributeNames =
                        getAttributeNames(edge).iterator();
                    attributeNames.hasNext(); ) {
                    String attributeName = (String)attributeNames.next();
                    subgraph.setAttribute(edge, attributeName,
                            getAttribute(edge, attributeName));
                }
            }
        } catch (DIFException e) {
            throw new RuntimeException(e.getMessage());
        } catch (IllegalActionException e) {
            throw new RuntimeException(e.getMessage());
        }
        return subgraph;
    }

    /** Returns true if the given object is a valid edge weight for this graph.
     *  @return True if the given object is a valid edge weight for this graph.
     */
    public boolean validEdgeWeight(Object object) {
        return object instanceof DIFEdgeWeight;
    }

    /** Returns true if the given object is a valid node weight for this graph.
     *  @return True if the given object is a valid node weight for this graph.
     */
    public boolean validNodeWeight(Object object) {
        return object instanceof DIFNodeWeight;
    }

    ///////////////////////////////////////////////////////////////////
    ////                       protected methods                   ////

    /** Reads an object and converts it to an appropriate token type, namely:
     *  String objects to String tokens, Double objects to Double tokens,
     *  double[][] to Double Matrix tokens and List objects to Array tokens.
     *  If the object is not one of these four types it is stored in an Object
     *  Token.
     *  @param token A token object.
     *  @return An object.
     *  @exception DIFException If a list or a double array is zero length.
     */
    protected static Token _convertObject2Token(Object object)
            throws DIFException {
        if(object instanceof Double) {
            return new DoubleToken(((Double)object).doubleValue());
        } else if(object instanceof String) {
            return new StringToken((String)object);
        } else if(object instanceof double[][]) {
            try {
                return new DoubleMatrixToken((double[][])object);
            } catch(IllegalActionException e) {
                throw new DIFException("Attempt to assign a null double"
                        +" matrix as an attribute.");
            }
        } else if(object instanceof List) {
            LinkedList tokenArray = new LinkedList();
            for(Iterator elements = ((List)object).iterator();
                elements.hasNext(); ) {
                Object element = elements.next();
                try {
                    tokenArray
                        .add(new ObjectToken(_convertObject2Token(element)));
                } catch(IllegalActionException e) {
                    throw new
                        RuntimeException("This exception should not happen.");
                }
            }
            try {
                return new
                    ArrayToken((Token[])tokenArray.toArray(new Token[0]));
            } catch(IllegalActionException e) {
                throw new DIFException("Attempt to assign a zero length"
                        +" List as an attribute.");
            }
        }
        try {
            return new ObjectToken(object);
        } catch(IllegalActionException e) {
            throw new RuntimeException("This exception should not happen.");
        }
    }

    /** Reads a token and converts it to an appropriate object type, namely:
     *  String tokens to String objects, Double tokens to Double objects,
     *  Double Matrix tokens to double[][], Array tokens to List objects and
     *  Object Tokens to objects.
     *  @param token A token object.
     *  @return An object. If the token type is not recognised returns null.
     */
    protected static Object _convertToken2Object(Token token) {
        if(token instanceof DoubleToken) {
            return new Double(((DoubleToken)token).doubleValue());
        } else if(token instanceof StringToken) {
            return ((StringToken)token).stringValue();
        } else if(token instanceof DoubleMatrixToken) {
            return ((DoubleMatrixToken)token).doubleMatrix();
        } else if(token instanceof ArrayToken) {
            LinkedList value = new LinkedList();
            for(int index = 0; index < ((ArrayToken)token).length(); index++) {
                Token element = (Token)
                    ((ObjectToken)((ArrayToken)token).getElement(index))
                    .getValue();
                value.add(_convertToken2Object(element));
            }
            return value;
        } else if(token instanceof ObjectToken) {
            return ((ObjectToken)token).getValue();
        }
        return null;
    }

    /** Register a new edge in the graph. The edge is assumed to
     *  be non-null, unique, and consistent with the node set.
     *  @exception IllegalArgumentException If the edge is unweighted or
     *  has a weight that is not an instance of {@link DIFEdgeWeight}.
     *  @param edge The new edge.
     */
    protected void _registerEdge(Edge edge) {
        if (!edge.hasWeight()) {
            throw new IllegalArgumentException("Attempt to add an unweighted "
                    + "edge to " + getClass().getName() + ".\n");
        } else if (!validEdgeWeight(edge.getWeight())) {
            throw new IllegalArgumentException("Invalid edge weight class: "
                    + edge.getWeight().getClass().getName() + ".\n"
                    + "A dump of the offending edge weight follows.\n"
                    + edge.getWeight().toString() + ".\n");
        } else {
            super._registerEdge(edge);
            _setAttributeContainer(edge, new NamedObj());
        }
    }

    /** Register a new node in the graph. The node is assumed to
     *  be non-null and unique.
     *  @param node The new node.
     *  @exception IllegalArgumentException If the node is unweighted or
     *  has a weight that is not an instance of {@link DIFNodeWeight}.
     */
    protected void _registerNode(Node node) {
        if (!node.hasWeight()) {
            throw new IllegalArgumentException("Attempt to add an unweighted "
                    + "node to " + getClass().getName() + ".\n");
        } else if (!validNodeWeight(node.getWeight())) {
            throw new IllegalArgumentException("Invalid node weight class: "
                    + node.getWeight().getClass().getName() + ".\n"
                    + "A dump of the offending node weight follows.\n"
                    + node.getWeight().toString() + ".\n");
        } else {
            super._registerNode(node);
            _setAttributeContainer(node, new NamedObj());
        }
    }

    ///////////////////////////////////////////////////////////////////
    ////                       private methods                     ////

    //  Returns a container in which Parameters representing the attributes of
    //  an edge, a node or the graph can be stored.<P>
    //  In setName method name of the returned container is used as the name
    //  of the edge/node or graph.
    //  @param object A graph edge/node or the graph itself.
    //  @return A container which contains all the parameters (if any) of this
    //  edge/node.
    //  @exception DIFException If the runtime class of this object
    //  is not an edge/node, it is not contained in the graph or it is not the
    //  graph itself.
    //
    protected NamedObj _getAttributeContainer(Object object)
            throws DIFException {
        if(object instanceof Edge) {
            if(!containsEdge((Edge)object)) {
                throw new DIFException("Attempt to get the"
                    + " attributes of an edge that is not in the graph."
                    + GraphException.edgeDump((Edge)object, this));
            }
            return (NamedObj)_containers.get(object);
        }

        if(object instanceof Node) {
            if(!containsNode((Node)object)) {
                throw new DIFException("Attempt to get the"
                    + " attributes of a node that is not in the graph."
                    + GraphException.nodeDump((Node)object, this));
            }
            return (NamedObj)_containers.get(object);
        }

        if(object == this) {
            return (NamedObj)_containers.get("graph");
        }

        throw new DIFException("Attempt to get the"
                + " attributes of an object whose runtime type is not"
                + " an edge/node in graph or which is not the"
                + " graph itself. Graph : " + getName()
                + "\nObject: " + object.toString());
    }

    // Sets a container for this object in which Parameters representing the
    //  attributes of an edge, a node or the graph can be stored. Returns the
    //  previous container.
    //  In setName method name of the returned container is used as the name
    //  of the edge/node or graph.
    //  @param object A graph edge/node or the graph itself.
    //  @param container A NamedObj.
    //
    protected NamedObj _setAttributeContainer(Object object,
            NamedObj container) {
        // No need to check if the object is in the graph or it is the
        // graph itself since this method is not being used by the user
        // even indirectly.
        if(object == this) {
            return (NamedObj)_containers.put("graph", container);
        }
        return (NamedObj)_containers.put(object, container);
    }

    ///////////////////////////////////////////////////////////////////
    ////                       private variables                   ////

    // This is a hash map that has nodes/edges or String("graph")
    // (representing this graph) as keys and NamedObj's as values.
    // NamedObj's are used as the containers of parameters of an object.
    //
    // Object(String("graph")/Edge/Node) -> NamedObj
    private HashMap _containers;

    // PS: Not using "graph" as a key to graph avoid errors because of
    // modifying a key.
}
