/* The class for data partitioning in dual homogeneous memory bank
 architectures.

 Copyright (c) 2002-2003 The University of Maryland. All rights reserved.
 Permission is hereby granted, without written agreement and without
 license or royalty fees, to use, copy, modify, and distribute this
 software and its documentation for any purpose, provided that the above
 copyright notice and the following two paragraphs appear in all copies
 of this software.

 IN NO EVENT SHALL THE UNIVERSITY OF MARYLAND BE LIABLE TO ANY PARTY
 FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
 THE UNIVERSITY OF MARYLAND HAS BEEN ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 THE UNIVERSITY OF MARYLAND SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 MARYLAND HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 ENHANCEMENTS, OR MODIFICATIONS.

*/

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

//////////////////////////////////////////////////////////////////////////
//// DataPartitioning
/** The class for data partitioning in dual homogeneous memory bank
architectures. It is developed for Mingyung's research. Basically, an
conflict graph should be provided. Generally, nodes representing data
variables while edges' meaning depends on applications.

@author Mingyung Ko
@version $Id: DataPartitioning.java,v 1.10 2003/04/21 18:26:25 myko Exp $
*/

public class DataPartitioning {

    /** A constructor with an conflict graph.
     *  @param conflictGraph The conflict graph to partition.
     */
    public DataPartitioning(PartitionedGraph conflictGraph) {
        _conflictGraph = conflictGraph;
        // Set two partitions as default.
        initPartitions(2);
    }

    ///////////////////////////////////////////////////////////////////
    ////                         public methods                    ////

    /** A two coloring algorithm. Two coloring is of polynomial complexity.
     *  @return The result as a {@link PartitionedGraph}.
     */
    public PartitionedGraph twoColoringStrategy() {
        initPartitions(2);
        Iterator components =
                _conflictGraph.connectedComponents().iterator();
        while (components.hasNext()) {
            Collection component = (Collection)components.next();
            Node seedNode = (Node)component.iterator().next();
            Graph graph = _conflictGraph.subgraph(component);
            // Two coloring for connected subgraphs.
            _twoColoringFor(seedNode, graph, 0);
            /* ---- random selection of seed partition ----
            int seedPartitionIndex = (Math.random() < 0.5) ? 0 : 1;
            _twoColoringFor(seedNode, graph, seedPartitionIndex);
            */
        }
        return _conflictGraph;
    }

    /** A greedy heuristic to solve data partitioning problems. The algorithm
     *  is based on the fact of sparse connection of conflict graphs.
     *  @param partitionCounts The number of partitions available.
     *  @return The result as a {@link PartitionedGraph}.
     */
    public PartitionedGraph greedyStrategy(int partitionCounts) {
        initPartitions(partitionCounts);
        // Sort connected components by decreasing size.
        Iterator downComponents = _conflictGraph.descendentListOf(
                _conflictGraph.connectedComponents()).iterator();
        while (downComponents.hasNext()) {
            Collection component = (Collection)downComponents.next();
            Node seed = (Node)
                    _conflictGraph.descendentListOf(component).get(0);
            _alternateAssignment(seed);
        }
        return _conflictGraph;
    }

    /** Initialize partitions by the given counts.
     *  @param counts Number of partitions to initialize.
     */
    public void initPartitions(int counts) {
        if (_conflictGraph.getPartitions().size() > 0)
            _conflictGraph.resetPartitions();
        for (int i = 0; i < counts; i++)
            _conflictGraph.addPartition(new GraphPartition());
    }

    /** Transform to OPBDP file format. OPBDP is an integer programming
     *  solver available on the site below.
     *  http://http://www.mpi-sb.mpg.de/units/ag2/software/opbdp/
     *  @return The format OPBDP accepts.
     */
    public String toOPBDPString() {
        String out = new String();
        String objective = new String();

        // compute total node values/weights.
        int total = 0;
        Iterator nodes = _conflictGraph.nodes().iterator();
        while (nodes.hasNext())
            total += _conflictGraph.valueOf(nodes.next());

        // optimization objective function (column variables)
        out += "min:";
        for (int i = 0; i < _conflictGraph.nodeCount(); i++) {
            Node node = _conflictGraph.node(i);
            int doubleValue = 2 * (int)_conflictGraph.valueOf(node);
            objective += " + " + doubleValue  + " C" + i;
        }
        objective += " - " + total;
        out += objective + ";\n\n";

        // constraint functions (row variables)
        int conflictCount = _conflictGraph.edgeCount();
        for (int i = 0; i < conflictCount; i++) {
            Edge edge = _conflictGraph.edge(i);
            Node node1 = edge.source();
            Node node2 = edge.sink();
            int node1Label = _conflictGraph.nodeLabel(node1);
            int node2Label = _conflictGraph.nodeLabel(node2);
            out += "R" + i + " : " +
                    "+ C" + node1Label + " + C" + node2Label + " = 1;\n";
        }
        out += "R" + conflictCount + " :" + objective + " >= 0;\n";
        return out;
    }

    ///////////////////////////////////////////////////////////////////
    ////                        private methods                    ////

    /*  Alternate partition assignment for a given node as seed.
     *  The assignment is for connected components only.
     *  @param seed The seed node.
     */
    private void _alternateAssignment(Node seed) {
        boolean assigned = false;
        List upPartitionList = _conflictGraph.ascendentPartitions();
        Iterator upPartitions = upPartitionList.iterator();
        while (upPartitions.hasNext()) {
            GraphPartition partition = (GraphPartition)upPartitions.next();
            if (!_linksExistFor(partition, seed)) {
                partition.addNode(seed);
                assigned = true;
                // The seed node is assigned properly. Quit the loop.
                break;
            }
        }
        // Certain cycles exist such that the connected component is not
        // bipartite. Some pair of variables can not be accessed in parallel.
        // They are assigned to the same bank and are accessed sequentially
        // one after the other.
        if (assigned == false) {
            // Assign the variable just to the bank with least usage so far.
            ((GraphPartition)upPartitionList.get(0)).addNode(seed);
        }
        _processNeighborsOf(seed);
    }

    /*  Check existence of links between the partition and node.
     *  @param partition The partition.
     *  @param node The node.
     *  @return True if there are links between them, false otherwise.
     */
    private boolean _linksExistFor(GraphPartition partition, Node node) {
        Iterator neighbors = _conflictGraph.neighbors(node).iterator();
        while (neighbors.hasNext()) {
            Node neighbor = (Node)neighbors.next();
            if (partition.containsNode(neighbor))
                return true;
        }
        return false;
    }

    /*  Bank assignment for neighbor nodes of a seed. This is to facilitate
     *  {@link #_alternateAssignment(Node)}.
     *  @param node The node whose neighbors are to be processed.
     */
    private void _processNeighborsOf(Node node) {
        Iterator neighbors =
                _conflictGraph.neighbors(node).iterator();
        while (neighbors.hasNext()) {
            Node neighbor = (Node)neighbors.next();
            if (_conflictGraph.partitionOf(neighbor) == null)
                _alternateAssignment(neighbor);
        }
    }

    /*  Evaluate the suitable partition to assign a node. This method is
     *  called by {@link greedyStrategy()}
     *  @param node A node to get the partition.
     *  @return The suitable partition to assign the node.
     */
/*
    private GraphPartition _suitablePartition(Node node) {
        Iterator upPartitions =
                _conflictGraph.ascendentPartitions().iterator();
        while (upPartitions.hasNext()) {
            GraphPartition partition = (GraphPartition)upPartitions.next();
            if (!_linksExistFor(partition, node))
                return partition;
        }
        return null;
    }
*/

    /*  Two coloring a node of a connected graph. Coloring is actually
     *  done by node assignment to partitions.
     *  @param node The node to color.
     *  @param graph The connected sub-graph.
     *  @param neighborIndex The partition index of the neighbor node.
     */
    private void _twoColoringFor(Node node, Graph graph, int neighborIndex) {
        if (_conflictGraph.partitionOf(node) == null) {
            int thisIndex;
            Collection neighborCollection = graph.neighbors(node);
            // Trivial components with only one node.
            if (neighborCollection.size() == 0) {
                ((GraphPartition)_conflictGraph.
                        getPartitions().get(0)).addNode(node);
                /* ---- Random selection of partition ----
                thisIndex = (Math.random() < 0.5) ? 0 : 1;
                ((GraphPartition)_conflictGraph.
                        getPartitions().get(thisIndex)).addNode(node);
                */

            } else {
                thisIndex = 1 - neighborIndex;
                ((GraphPartition)_conflictGraph.
                        getPartitions().get(thisIndex)).addNode(node);
                Iterator neighbors = neighborCollection.iterator();
                while (neighbors.hasNext()) {
                    Node neighbor = (Node)neighbors.next();
                    _twoColoringFor(neighbor, graph, thisIndex);
                }
            }
        }
    }

    ///////////////////////////////////////////////////////////////////
    ////                       private variables                   ////

    /* The conflict graph. */
    private PartitionedGraph _conflictGraph;
}
