/* A base scheduler using multirate cycle breaking strategies.

Copyright (c) 2002 The University of Maryland.

Permission is hereby granted, without written agreement and without
license or royalty fees, to use, copy, modify, and distribute this
software and its documentation for any purpose, provided that the above
copyright notice and the following two paragraphs appear in all copies
of this software.

IN NO EVENT SHALL THE UNIVERSITY OF MARYLAND BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
THE UNIVERSITY OF MARYLAND HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

THE UNIVERSITY OF MARYLAND SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
MARYLAND HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS.

@AcceptedRating Red

*/


import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;

import ptolemy.actor.sched.Firing;
import ptolemy.actor.sched.Schedule;
import ptolemy.actor.sched.ScheduleElement;

//////////////////////////////////////////////////////////////////////////
//// MultirateCycleBreaking
/** A base scheduler using multirate cycle breaking strategies.
The scheduler computes a schedule by removing edges. The removal of
edges breaks cycles or strongly connected components of a graph. With
the breaking, result schedules can be evaluated by easier acyclic
schedulers.
<p>
In this class, {@link #_cycleBreaking(SDFGraph)},
{@link #_acyclicScheduling(SDFGraph, int)},
{@link #_tightScheduling(SDFGraph, int)}, and
{@link #_computeSchedule(SDFGraph, int)} are to be re-defined for
various algorithms respectively. This base class implements edge
removal with largest normalized delays (delay / TNSE). Generic
topological sorting is implemented in acyclic scheduling while
nothing is done in tight scheduling.

@author Mingyung Ko
@version $Id: MultirateCycleBreaking.java,v 1.12 2002/09/23 17:22:25 shahrooz Exp $
*/

public class MultirateCycleBreaking extends ClusterScheduler {

    /** Constructor for a given SDF graph.
     *
     *  @param graph The given SDF graph.
     */
    public MultirateCycleBreaking(SDFGraph graph) {
        super(graph);
        _removedEdgesMap = new HashMap();
    }

    ///////////////////////////////////////////////////////////////////
    ////                       protected methods                   ////

    /** Acyclic scheduling for the given graph. The graph
     *  must be acyclic and assumes at least 2 nodes.
     *  <p>
     *  The implementation is just a generic topological sorted and
     *  linear schedule. Users can redefine this method.
     *
     *  @param graph The given graph in {@link SDFGraph}.
     *  @param graphRepetitions Repetitions for the given graph.
     *  @return The acyclic schedule.
     */
    protected ScheduleElement _acyclicScheduling(
            SDFGraph graph, int graphRepetitions) {

        // Initialize a new schedule for the graph.
        Schedule schedule = new Schedule();
        schedule.setIterationCount(graphRepetitions);

        // Compute topological sorting as the acyclic schedule.
        DirectedAcyclicGraph acyclicGraph = graph.toDirectedAcyclicGraph();
        Object[] sortedWeights = acyclicGraph.topologicalSort();
        for (int i = 0; i < sortedWeights.length; i++) {
            Node node = graph.node(sortedWeights[i]);
            int repetitions = graph.getRepetitions(node);

            // Expand super nodes and compute associated subgraphs.
            if (_isSuperNode(node)) {
                SDFGraph subgraph = _getSubgraph(node);
                schedule.add(_computeSchedule(subgraph, repetitions));
            // Get a firing from SDFSchedule.
            } else {
                Firing firing = _sdfSchedule.newFiring(node);
                firing.setIterationCount(repetitions);
                schedule.add(firing);
            }
        }
        return schedule;
    }

    /** Compute the schedule.
     *
     *  @return The result {@link SDFSchedule}.
     */
    protected Object _compute() {
        _toSDFSchedule(
                _computeSchedule(_getClusteredGraph(), 1), _sdfSchedule);
        return _sdfSchedule;
    }

    /** Compute the schedule for the given SDF graph. The method handles
     *  the special case of one-node graph.
     *
     *  @param graph The given graph.
     *  @param graphRepetitions Repetitions for the given graph.
     *  @return The result Ptolemy schedule.
     */
    protected ScheduleElement _computeSchedule(
            SDFGraph graph, int graphRepetitions) {

        ScheduleElement schedule = null;

        // The graph has only one (atomic, not super) node.
        if (graph.nodeCount() == 1) {
            Node node = (Node)graph.nodes().iterator().next();
            schedule = _sdfSchedule.newFiring(node);
            schedule.setIterationCount(graphRepetitions);
        // Graphs with at least 2 nodes.
        } else {
            DirectedGraph[] sccs = graph.sccDecomposition();

            // The graph is strongly connected.
            if (sccs.length == 1) {
                do {
                    Collection edgesToRemove = _cycleBreaking(graph);
                    // No edges to break. The graph is tight.
                    if (edgesToRemove.isEmpty()) {
                        _putBackEdges(graph);
                        return _tightScheduling(graph, graphRepetitions);

                    // Remove an edge with maximum normalized delay.
                    } else {
                        Edge edge = (Edge)edgesToRemove.iterator().next();
                        ArrayList edgeList = new ArrayList();
                        edgeList.add(edge);
                        _removeEdges(graph, edgeList);
                    }
                    sccs = graph.sccDecomposition();
                } while (sccs.length == 1);
            }

            // The graph must be SCC partitionable.
            for (int i = 0; i < sccs.length; i++)
                _clusterNodes(graph, sccs[i].nodes(),
                        new Node(new SDFNodeWeight()));
            schedule = _acyclicScheduling(graph, graphRepetitions);
        }
        return schedule;
    }

    /** Cycle breaking rule for the graph. The candidate edges
     *  to remove to break cycles are return as a collection.
     *  However, these edges are not actually removed from graph.
     *  To remove edges, please use
     *  {@link #_removeEdges(SDFGraph, Collection)}.
     *  <p>
     *  The implementation here is to select the edge with
     *  largest normalized delays (delay / TNSE). However, users
     *  can redefine this method.
     *
     *  @param graph The given {@link SDFGraph} to break cycles.
     *  @return The collection of edges to remove to break cycles.
     */
    protected Collection _cycleBreaking(SDFGraph graph) {
        double maxNormalizedDelay = 0.0;
        Edge maxDelayEdge = null;

        // Compute and pick the edge with largest delays.
        Iterator edges = graph.edges().iterator();
        while (edges.hasNext()) {
            Edge edge = (Edge)edges.next();
            int delay = ((SDFEdgeWeight)edge.getWeight()).getDelay();
            double normalizedDelay =
                    (double)delay / (double)graph.TNSE(edge);
            if (normalizedDelay > maxNormalizedDelay) {
                maxNormalizedDelay = normalizedDelay;
                maxDelayEdge = edge;
            }
        }

        ArrayList edgesToRemove = new ArrayList();
        if (maxNormalizedDelay >= 1.0)
            edgesToRemove.add(maxDelayEdge);
        return edgesToRemove;
    }

    /** Test the graph's tight inter-dependency.
     *
     *  @graph The given {@link SDFGraph}.
     *  @return True if the graph is tightly interdependent.
     */
    protected boolean _isTight(SDFGraph graph) {
        SDFGraph cloneGraph = (SDFGraph)graph.clone();
        ArrayList bigDelayEdges = new ArrayList();

        // Collect edges with sufficient delays.
        Iterator edges = cloneGraph.edges().iterator();
        while (edges.hasNext()) {
            Edge edge = (Edge)edges.next();
            int delay = ((SDFEdgeWeight)edge.getWeight()).getDelay();
            if (delay >= cloneGraph.TNSE(edge))
                bigDelayEdges.add(edge);
        }
        // Remove edges with big delays.
        edges = bigDelayEdges.iterator();
        while (edges.hasNext())
            cloneGraph.removeEdge((Edge)edges.next());

        // Graph is tight if it's cyclic after edge removal.
        return !cloneGraph.isAcyclic();
    }

    /** Edges to put back into graph. This method is usually
     *  executed before tight scheduling.
     *
     *  @param graph The given {@link SDFGraph}.
     *  @param edges The edges to put back.
     */
    final protected void _putBackEdges(SDFGraph graph) {
        if (!_removedEdgesMap.containsKey(graph))
            return;

        Iterator edgesToPutBack =
                ((Collection)_removedEdgesMap.get(graph)).iterator();
        while (edgesToPutBack.hasNext())
            graph.addEdge((Edge)edgesToPutBack.next());
        // Remove the map entry from _removedEdgesMap.
        _removedEdgesMap.remove(graph);

    }

    /** Remove edges from the given graph. The removed edges are
     *  kept in <code>_removedEdgesMap</code> for possibe
     *  future putting back.
     *
     *  @param graph The given {@link SDFGraph}.
     *  @param edgeCollection The edges to remove.
     */
    final protected void _removeEdges(
            SDFGraph graph, Collection edgeCollection) {

        /* Record removed edges in _removedEdgesMap. */
        // Remove edges from graphs not existent in _removedEdgesMap.
        if (!_removedEdgesMap.containsKey(graph)) {
            HashSet edgesToRecord = new HashSet();
            edgesToRecord.addAll(edgeCollection);
            _removedEdgesMap.put(graph, edgesToRecord);
        // Remove edges from graphs existent in _removedEdgesMap.
        } else {
            HashSet edgesExistent = (HashSet)_removedEdgesMap.get(graph);
            edgesExistent.addAll(edgeCollection);
        }

        // Actual edges removal from graphs.
        Iterator edges = edgeCollection.iterator();
        while (edges.hasNext())
            graph.removeEdge((Edge)edges.next());
    }

    /** Tight scheduling for the given graph. The method assumes
     *  an input graph with at least 2 nodes.
     *  <p>
     *  <em>FIXME</em>: There is no implementation so far and
     *  users should define this method.
     *
     *  @param graph The given {@link SDFGraph}.
     *  @param graphRepetitions Repetitions for the given graph.
     *  @return The tight schedule.
     */
    protected ScheduleElement _tightScheduling(
            SDFGraph graph, int graphRepetitions) {
        throw new RuntimeException("Not yet implemented.");
    }

    ///////////////////////////////////////////////////////////////////
    ////                       private variables                   ////

    /* A map from graphs to the removed edges. */
    private HashMap _removedEdgesMap;
}
