/* A base class for partitioned graph structure.

 Copyright (c) 2002 The University of Maryland. All rights reserved.
 Permission is hereby granted, without written agreement and without
 license or royalty fees, to use, copy, modify, and distribute this
 software and its documentation for any purpose, provided that the above
 copyright notice and the following two paragraphs appear in all copies
 of this software.

 IN NO EVENT SHALL THE UNIVERSITY OF MARYLAND BE LIABLE TO ANY PARTY
 FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
 THE UNIVERSITY OF MARYLAND HAS BEEN ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 THE UNIVERSITY OF MARYLAND SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 MARYLAND HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 ENHANCEMENTS, OR MODIFICATIONS.

*/

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;

//////////////////////////////////////////////////////////////////////////
//// PartitionBase
/** A base class for partitioned graph structure. This class provides
commonly used methods for
{@link GraphPartition} and {@link PartitionedGraph}.
Basically, this class is still an instance of graph because all partition
structures are indeed graphs.

@author Mingyung Ko
@version $Id: PartitionBase.java,v 1.3 2002/12/24 23:07:43 keceli Exp $
*/

public class PartitionBase extends Graph {

    /** A constructor with node and edge values. Definition of values
     *  totally depends on the application. However, they should be
     *  numbers and 'double' is chosen in a most general sense.
     *  @param nodeValues Values associated to nodes.
     *  @param edgeValues Values associated to edges.
     */
    public PartitionBase() {}

    ///////////////////////////////////////////////////////////////////
    ////                         public methods                    ////

    /** An ascendently ordered list of objects. It is the associated
     *  values, see {@link #valueOf(Object)}, being sorted. The
     *  argument comes from nodes()/edges() methods in class {@link Graph}
     *  or a collection of {@link GraphPartition}s.
     *  @param objects The objects to be ordered.
     *  @return An ascendent list of the objects.
     */
    public List ascendentListOf(Collection objects) {
        ArrayList objectList = new ArrayList(objects);
        Collections.sort(objectList, new upComparator());
        return objectList;
    }

    /** A descendently ordered list of objects. It is the associated
     *  values, see {@link #valueOf(Object)}, being sorted. The
     *  argument comes from nodes()/edges() methods in class {@link Graph}
     *  or a collection of {@link GraphPartition}s.
     *  @param elements The objects to be ordered.
     *  @return A descendent list of the objects.
     */
    public List descendentListOf(Collection objects) {
        ArrayList objectList = new ArrayList(objects);
        Collections.sort(objectList, new downComparator());
        return objectList;
    }

    /** Get the value associated with the object. Valid objects are
     *  node, edge, collection of nodes/edges, and graph partition.
     *  Validity checking is enforced by throwing exceptions.
     *  For graph partitions, the total value
     *  of NODEs is returned.
     *  @param object The object to get the associated value.
     *  @return The associated value.
     *  @exception IllegalArgumentException If object invalid to get value.
     */
    public double valueOf(Object object) {
        if (object instanceof Element) {
            _checkGraphElement((Element)object);
            return ((Double)((Element)object).getWeight()).doubleValue();

        } else if (object instanceof GraphPartition) {
            // Nodes are chosen to amount up their values.
            return valueOf(((GraphPartition)object).nodes());

        } else if (object instanceof Collection) {
            double total = 0.0;
            Iterator objects = ((Collection)object).iterator();
            while (objects.hasNext())
                total += valueOf(objects.next());
            return total;

        } else {
            throw new IllegalArgumentException(
                    "Invalid object to get value.");
        }
    }

    ///////////////////////////////////////////////////////////////////
    ////                        private methods                    ////

    /*  Check existence in the graph for a given element.
     *  @param element The element to check.
     *  @exception GraphElementException If the element is nonexistent.
     */
    private void _checkGraphElement(Element element) {
        if (element instanceof Node)
            GraphElementException.checkNode((Node)element, this);
        else
            GraphElementException.checkEdge((Edge)element, this);
    }

    ///////////////////////////////////////////////////////////////////
    ////                       private variables                   ////

    // Ascendently compare partition's total value.
    private class upComparator implements Comparator {
        public int compare(Object object1, Object object2) {
            return (int)(valueOf(object1) - valueOf(object2));
        }
    }

    // Descendently compare partition's total value.
    private class downComparator implements Comparator {
        public int compare(Object object1, Object object2) {
            return (int)(valueOf(object2) - valueOf(object1));
        }
    }
}
