/* A graph with partitions.

 Copyright (c) 2002 The University of Maryland. All rights reserved.
 Permission is hereby granted, without written agreement and without
 license or royalty fees, to use, copy, modify, and distribute this
 software and its documentation for any purpose, provided that the above
 copyright notice and the following two paragraphs appear in all copies
 of this software.

 IN NO EVENT SHALL THE UNIVERSITY OF MARYLAND BE LIABLE TO ANY PARTY
 FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
 THE UNIVERSITY OF MARYLAND HAS BEEN ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 THE UNIVERSITY OF MARYLAND SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 MARYLAND HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 ENHANCEMENTS, OR MODIFICATIONS.

*/

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

//////////////////////////////////////////////////////////////////////////
//// PartitionedGraph
/** A graph of partitions with additional partition related tools. The
partitions are not necessarily disjoint. Disjointness is up to the uers
to maintain.

@author Mingyung Ko
@version $Id: PartitionedGraph.java,v 1.4 2002/12/24 23:07:43 keceli Exp $
*/

public class PartitionedGraph extends PartitionBase {

    /** A constructor with node and edge values. Definition of values
     *  totally depends on the application. However, they should be
     *  numbers and 'double' is chosen in a most general sense.
     *  @param nodeValues Values associated to nodes.
     *  @param edgeValues Values associated to edges.
     */
    public PartitionedGraph() {
        resetPartitions();
    }

    /** Constructor for a given graph. Construct the partitioned graph by
     *  duplicating all nodes and edges.
     *  @param graph The graph to construct from.
     */
    public PartitionedGraph(Graph graph) {
        resetPartitions();
        addNodes(graph.nodes());
        addEdges(graph.edges());
    }

    ///////////////////////////////////////////////////////////////////
    ////                         public methods                    ////

    /** Add the given graph partition.
     *  @param partition The graph partition to add.
     */
    public void addPartition(GraphPartition partition) {
        _partitions.add(partition);
    }

    /** An ascendently ordered list of partitions. It is the associated
     *  total values, see {@link GraphPartition#totalValue()},
     *  being sorted.
     *  @return An ascendent list of partitions.
     */
    public List ascendentPartitions() {
        return ascendentListOf(getPartitions());
    }

    /** A descendently ordered list of partitions. It is the associated
     *  total values, see {@link GraphPartition#totalValue()},
     *  being sorted.
     *  @return A descendent list of partitions.
     */
    public List descendentPartitions() {
        return descendentListOf(getPartitions());
    }

    /** Get the edge cut set over all partitions. See also
     *  {@link edgeCutOf(Collection)}.
     *  @return The set of edge cut.
     */
    public Collection edgeCut() {
        return edgeCutOf(getPartitions());
    }

    /** Get edge cut set over a collection of partitions. Edge cut is a
     *  set of edges with ending nodes in distinct partitions.
     *  @param partitionCollection A collection of partitions.
     *  @return The set of edge cut.
     */
    public Collection edgeCutOf(Collection partitionCollection) {
        ArrayList cut = new ArrayList();
        Iterator edges = edges().iterator();
        while (edges.hasNext()) {
            Edge edge = (Edge)edges.next();
            GraphPartition sourcePartition = partitionOf(edge.source());
            GraphPartition sinkPartition = partitionOf(edge.sink());
            if (partitionCollection.contains(sourcePartition) &&
                    partitionCollection.contains(sinkPartition)) {
                if (sourcePartition != sinkPartition)
                    cut.add(edge);
            }
        }
        return cut;
    }

    /** Get the edge cut set of two partitions. See also
     *  {@link #edgeCutOf(Collection)}.
     *  @param p1 The first partition.
     *  @param p2 The second partition.
     *  @return The set of edge cut.
     */
    public Collection edgeCutOf(GraphPartition p1, GraphPartition p2) {
        ArrayList partitions = new ArrayList();
        partitions.add(p1);
        partitions.add(p2);
        return edgeCutOf(partitions);
    }

    /** Display edge cut in a text string.
     *  @param edgeCut The edge cut to print.
     *  @return The output.
     */
    public String edgeCutString(Collection edgeCut) {
        String out = new String();
        Iterator cuts = edgeCut.iterator();
        while (cuts.hasNext()) {
            Edge cut = (Edge)cuts.next();
            Node n1 = cut.source();
            Node n2 = cut.sink();
            int p1Index = getPartitions().indexOf(partitionOf(n1));
            int p2Index = getPartitions().indexOf(partitionOf(n2));
            out += "edge cut set :\n";
            out += "(node" + nodeLabel(n1) + ", value" + valueOf(n1) +
                    ", part" + p1Index + ") : (" +
                    "node" + nodeLabel(n2) + ", value" + valueOf(n2) +
                    ", part" + p2Index + ")\n";
        }
        return out;
    }

    /** Exchange container partitions for the given nodes. The order of
     *  arguments does not matter and same result will be produced.
     *  @param n1 The first node.
     *  @param n2 The second node.
     */
    public void exchangePartitionsFor(Node n1, Node n2) {
    }

    /** Get all the partitions. The partition list returned is not allowed
     *  to modify. This is to ensure that partition additions are through
     *  {@link addPartition(GraphPartition)}.
     *  @return Present partitions in type of <code>List</code>.
     */
    public List getPartitions() {
        return _partitions;
    }

    /** Get the container partition of an element.
     *  @param element The given element.
     *  @return The container partition.
     */
    public GraphPartition partitionOf(Element element) {
        List partitions = getPartitions();
        for (int i = 0; i < partitions.size(); i++) {
            GraphPartition partition = (GraphPartition)partitions.get(i);
            if ( (element instanceof Node) &&
                    partition.containsNode((Node)element) ) {
                return partition;

            } else if ( (element instanceof Edge) &&
                    partition.containsEdge((Edge)element) ) {
                return partition;
            }
        }
        return null;
    }

    /** Display partitions in a text string.
     *  @return The output.
     */
    public String partitionString() {
        String out = new String();
        int partitionCounts = getPartitions().size();
        out += "partition counts = " + partitionCounts + "\n\n";
        for (int i = 0; i < partitionCounts; i++) {
            GraphPartition partition =
                    (GraphPartition)getPartitions().get(i);
            out += "partition " + i +
                    " of total value " + valueOf(partition) + "\n";
            Iterator nodes = partition.nodes().iterator();
            while (nodes.hasNext()) {
                Node node = (Node)nodes.next();
                out += "node" + nodeLabel(node) +
                        ", value" + valueOf(node) + "\n";
            }
            out += "\n";
        }
        return out;
    }

    /** Remove all partitions.
     */
    public void resetPartitions() {
        _partitions = new ArrayList();
    }

    ///////////////////////////////////////////////////////////////////
    ////                        private methods                    ////

    ///////////////////////////////////////////////////////////////////
    ////                       private variables                   ////

    // The list of partitions.
    private List _partitions;

}
