/* Information associated with an SDF edge.

 Copyright (c) 2001-2002 The University of Maryland.
 All rights reserved.
 Permission is hereby granted, without written agreement and without
 license or royalty fees, to use, copy, modify, and distribute this
 software and its documentation for any purpose, provided that the above
 copyright notice and the following two paragraphs appear in all copies
 of this software.

 IN NO EVENT SHALL THE UNIVERSITY OF MARYLAND BE LIABLE TO ANY PARTY
 FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
 THE UNIVERSITY OF MARYLAND HAS BEEN ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 THE UNIVERSITY OF MARYLAND SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 MARYLAND HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 ENHANCEMENTS, OR MODIFICATIONS.

*/


import java.lang.reflect.Array;
import ptolemy.actor.IOPort;

//////////////////////////////////////////////////////////////////////////
//// SDFEdgeWeight
/** Information associated with an SDF edge.
SDFEdgeWeights are objects associated with {@link ptolemy.graph.Edge}s
that represent SDF edges in {@link ptolemy.graph.Graph}s.
This class caches frequently-used data associated with SDF edges.
It is also useful for intermediate SDF graph representations
that do not correspond to Ptolemy II models, and performing graph
transformations (e.g., vectorization and retiming).
It is intended for use with analysis/synthesis algorithms that operate
on generic graph representations of SDF models.

@author Michael Rinehart, Shuvra S. Bhattacharyya, Ming-Yung Ko
@version $Id: SDFEdgeWeight.java,v 1.2 2002/07/16 18:29:16 myko Exp $
@see ptolemy.graph.Edge
*/

public class SDFEdgeWeight extends CSDFEdgeWeight {

    /** Construct an edge weight for a homogeneous, zero-delay edge.
     */
    public SDFEdgeWeight() {
        setSinkPort(null);
        setSourcePort(null);
        setDelay(0);
        setSDFProductionRate(1);
        setSDFConsumptionRate(1);
    }

    /** Construct an edge weight for a specified token production rate,
     *  token consumption rate, and delay.
     *  @param productionRate The token production rate.
     *  @param consumptionRate The token consumption rate.
     *  @param delay The delay.
     */
    public SDFEdgeWeight(int productionRate, int consumptionRate, int delay) {
        setSourcePort(null);
        setSinkPort(null);
        setDelay(delay);
        setSDFProductionRate(productionRate);
        setSDFConsumptionRate(consumptionRate);
    }

    /** Construct an edge weight for a specified source port, sink port,
     *  token production rate, token consumption rate, and delay.
     *  The source port and sink port are ports (in a Ptolemy II SDF
     *  model) that correspond to this edge.
     *  @param sourcePort The source port.
     *  @param sinkPort The sink port.
     *  @param productionRate The token production rate.
     *  @param consumptionRate The token consumption rate.
     *  @param delay The delay.
     */
    public SDFEdgeWeight(IOPort sourcePort, IOPort sinkPort,
           int productionRate, int consumptionRate, int delay) {

        setSinkPort(sinkPort);
        setSourcePort(sourcePort);
        setDelay(delay);
        setSDFProductionRate(productionRate);
        setSDFConsumptionRate(consumptionRate);
    }

    ///////////////////////////////////////////////////////////////////
    ////                         public methods                    ////

    /** Get the token consumption rate of the associated SDF edge.
     *  @return The consumption rate.
     */
    public int getSDFConsumptionRate() {
        return getCSDFConsumptionRate(0);
    }

    /** Get the token production rate of the associated SDF edge.
     *  @return The production rate.
     */
    public int getSDFProductionRate() {
        return getCSDFProductionRate(0);
    }

    /** Set the token consumption rate of the SDF edge.
     *  @param consumptionRate The token consumption rate in
     *  <code>int</code>.
     */
    public void setSDFConsumptionRate(int consumptionRate) {
        int[] rates = {consumptionRate};
        setCSDFConsumptionRates(rates);
    }

    /** Set the token production rate of the SDF edge.
     *  @param productionRate The new token production rate.
     */
    public void setSDFProductionRate(int productionRate) {
        int[] rates = {productionRate};
        setCSDFProductionRates(rates);
    }

    /** Return a string representation of the edge weight. This string
     *  representation is in the following form:
     *  <p>
     *  <em> productionRate consumptionRate delay </em>
     *  <p>
     * @return The string representation of the edge weight.
     */
    public String toString() {
        return new String(getSDFProductionRate() + " " +
                getSDFConsumptionRate() + " " + getDelay());
    }

}
