/* Information associated with an SDF graph.

 Copyright (c) 2001-2002 The University of Maryland. All rights reserved.
 Permission is hereby granted, without written agreement and without
 license or royalty fees, to use, copy, modify, and distribute this
 software and its documentation for any purpose, provided that the above
 copyright notice and the following two paragraphs appear in all copies
 of this software.

 IN NO EVENT SHALL THE UNIVERSITY OF MARYLAND BE LIABLE TO ANY PARTY
 FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
 THE UNIVERSITY OF MARYLAND HAS BEEN ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 THE UNIVERSITY OF MARYLAND SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 MARYLAND HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 ENHANCEMENTS, OR MODIFICATIONS.

*/


import java.util.Iterator;

import ptolemy.math.Fraction;


//////////////////////////////////////////////////////////////////////////
//// SDFGraph
/** Information associated with an SDF graph.
This class caches frequently-used data associated with SDF graphs.
It is also useful for intermediate SDF graph representations
(between the application model and implementation), and performing graph
transformations (e.g. convertion to and manipulation of single-rate graphs).
It is intended for use with analysis/synthesis algorithms that operate
on generic graph representations of SDF models.
<p>
SDFGraph nodes and edges have weights of type {@link SDFNodeWeight} and
{@link SDFEdgeWeight}, respectively.

@author Ming-Yung Ko, Shuvra S. Bhattacharyya
@version $Id: SDFGraph.java,v 1.6 2002/09/18 16:05:52 myko Exp $
@see SDFEdgeWeight
@see SDFNodeWeight
*/

public class SDFGraph extends CSDFGraph {

    /** Construct an empty SDF graph.
     */
    public SDFGraph() {
        super();
    }

    /** Construct an empty SDF graph with enough storage allocated for the
     *  specified number of nodes.
     *  @param nodeCount The number of nodes.
     */
    public SDFGraph(int nodeCount) {
        super(nodeCount);
    }

    /** Construct an empty SDF graph with enough storage allocated for the
     *  specified number of edges, and number of nodes.
     *  @param nodeCount The integer specifying the number of nodes
     *  @param edgeCount The integer specifying the number of edges
     */
    public SDFGraph(int nodeCount, int edgeCount) {
        super(nodeCount, edgeCount);
    }

    ///////////////////////////////////////////////////////////////////
    ////                         public methods                    ////

    /** <em>Buffer Memory Lower Bound</em> of the graph. It is derived
     *  from the sum of all edge's BMLB value.
     *  @return The lower bound for whole graph.
     */
    public int BMLB() {
        int bound = 0;
        Iterator edges = edges().iterator();
        while (edges.hasNext()) {
            bound += BMLB((Edge)edges.next());
        }
        return bound;
    }

    /** <em>Buffer Memory Lower Bound</em> of a given edge.
     *  @param edge The given edge.
     *  @return The lower bound.
     */
    public int BMLB(Edge edge) {
        SDFEdgeWeight weight = (SDFEdgeWeight)edge.getWeight();
        int production  = weight.productionPeriodRate();
        int consumption = weight.consumptionPeriodRate();
        int delay       = weight.getDelay();
        int divisor = Fraction.gcd(production, consumption);
        int tokenExchange = production * consumption / divisor;
        if (delay < tokenExchange)
            return (delay + tokenExchange);
        else
            return delay;
    }

    /** <em>Buffer Memory Upper Bound</em> of the graph. It is derived
     *  from the sum of all edge's BMLB value.
     *  @return The lower bound for whole graph.
     */
    public int BMUB() {
        int bound = 0;
        Iterator edges = edges().iterator();
        while (edges.hasNext()) {
            bound += BMUB((Edge)edges.next());
        }
        return bound;
    }

    /** <em>Buffer Memory Upper Bound</em> of a given edge.
     *  @param edge The given edge.
     *  @return The upper bound.
     */
    public int BMUB(Edge edge) {
        int delay = ((SDFEdgeWeight)edge.getWeight()).getDelay();
        return (delay + TNSE(edge));
    }

    /** "Total Number of Samples Exchanged" of a given edge. The value
     *  is equal to product of sink node repetitions and consumption
     *  rate, or product of source node repetitions and production rate.
     *
     *  @param edge The given edge.
     *  @return The number of samples exchanged.
     */
    public int TNSE(Edge edge) {
        int sinkRepetitions = getRepetitions(edge.sink());
        SDFEdgeWeight edgeWeight = (SDFEdgeWeight)edge.getWeight();
        int consumptionRate = edgeWeight.getSDFConsumptionRate();
        return (sinkRepetitions * consumptionRate);
    }

    /** Verify edge weight for SDF graph.
     *  @param weight The edge weight to verify.
     *  @return True if the given edge weight is valid for SDF graph.
     */
    public boolean validEdgeWeight(Object weight) {
        return weight instanceof SDFEdgeWeight;
    }

    /** Verify node weight for SDF graph.
     *  @param weight The node weight to verify.
     *  @return True if the given node weight is valid for SDF graph.
     */
    public boolean validNodeWeight(Object weight) {
        return weight instanceof SDFNodeWeight;
    }
}
