/* A base SDF scheduler

 Copyright (c) 2001-2002 The University of Maryland. All rights reserved.
 Permission is hereby granted, without written agreement and without
 license or royalty fees, to use, copy, modify, and distribute this
 software and its documentation for any purpose, provided that the above
 copyright notice and the following two paragraphs appear in all copies
 of this software.

 IN NO EVENT SHALL THE UNIVERSITY OF MARYLAND BE LIABLE TO ANY PARTY
 FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
 THE UNIVERSITY OF MARYLAND HAS BEEN ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 THE UNIVERSITY OF MARYLAND SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 MARYLAND HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 ENHANCEMENTS, OR MODIFICATIONS.

*/


import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import ptolemy.actor.sched.Firing;
import ptolemy.actor.sched.Schedule;
import ptolemy.actor.sched.ScheduleElement;

//////////////////////////////////////////////////////////////////////////
//// SDFScheduler
/** A base class for SDF schedulers. This class extends
{@link ptolemy.graph.analysis.Analysis}. The advantage of this inheritance
is that schedule re-computation is necessary on if the graph is obsolete.

@author Mingyung Ko
@version $Id: SDFScheduler.java,v 1.7 2002/09/20 16:49:18 myko Exp $
*/

public abstract class SDFScheduler extends CSDFScheduler {

    /** Constructor for a given graph.
     *
     *  @param graph The given SDF graph.
     */
    public SDFScheduler(SDFGraph graph) {
        super(graph);
        _sdfSchedule = new SDFSchedule();
    }

    ///////////////////////////////////////////////////////////////////
    ////                         public methods                    ////

    /** Get buffer living history. The history is saved in the form of
     *  Iterator. Each Iterator element is a map from living buffers/edges
     *  to result tokens after Firing execution.
     *  The order of Iterator is the same as
     *  {@link SDFSchedule#firingIterator()}. It means that the referenced
     *  time is based on the execution of Firings. Therefore, this method
     *  is costly due to simulating the real execution.
     *
     *  @param graph The given SDF graph.
     *  @param schedule The given SDF schedule.
     *  @return An Iterator of Maps from living buffers/edge to tokens.
     */
    static public Iterator bufferHistory(
            SDFGraph graph, SDFSchedule schedule) {

        int totalBuffers = graph.edgeCount();
        int[] tokens = new int[totalBuffers];
        ArrayList history = new ArrayList();

        // Set initial buffer cost to the number of delays.
        Iterator edges = graph.edges().iterator();
        while (edges.hasNext()) {
            Edge edge = (Edge)edges.next();
            int edgeLabel = graph.edgeLabel(edge);
            tokens[edgeLabel] = ((SDFEdgeWeight)edge.getWeight()).getDelay();
        }

        // Flatten the firing execution sequence and simulate
        Iterator firings = schedule.firingIterator();
        while (firings.hasNext()) {
            Firing firing = (Firing)firings.next();
            Node node = schedule.getNode(firing);
            int count = firing.getIterationCount();

            // For this Firing, check all buffers linked to it.
            Map livingBuffers = new HashMap();
            Iterator incidentEdges = graph.incidentEdges(node).iterator();
            while (incidentEdges.hasNext()) {
                Edge edge = (Edge)incidentEdges.next();
                int edgeLabel = graph.edgeLabel(edge);
                SDFEdgeWeight weight = (SDFEdgeWeight)edge.getWeight();

                // Save token results.
                if (node == edge.source()) {
                    int production = weight.getSDFProductionRate();
                    tokens[edgeLabel] += production * count;
                } else {
                    int consumption = weight.getSDFConsumptionRate();
                    tokens[edgeLabel] -= consumption * count;
                }

                // Buffers are alive due to execution of this Firing.
                livingBuffers.put(edge, new Integer(tokens[edgeLabel]));
            }

            // Buffers are alive due to deposit of tokens.
            for (int i = 0; i < totalBuffers; i++) {
                Edge edge = graph.edge(i);
                if ( (tokens[i] > 0) & (!livingBuffers.containsKey(edge)) )
                    livingBuffers.put(edge, new Integer(tokens[i]));
            }

            // Save buffers living during this Firing execution.
            history.add(livingBuffers);
        }
        return history.iterator();
    }

    /** Build buffer intersections(lifetime overlaps). Buffer intersections
     *  are useful information to buffer sharing. The intersections are saved
     *  in the form of Map of Sets. Each map entry maps an edge to a set of
     *  edges intersecting with it. Execution simulation used here to collect
     *  buffer intersections is a universal but costly strategy. Descendent
     *  classes may override this method with more efficient and clever
     *  algorithms.
     *
     *  @return The buffer intersections in the form of Map of Sets.
     */
    public Map bufferIntersections() {
        return bufferIntersections(
                (SDFGraph)graph(), (SDFSchedule)schedule());
    }

    /** Build buffer intersections(lifetime overlaps). Buffer intersections
     *  are useful information to buffer sharing. The intersections are saved
     *  in the form of Map of Sets. Each map entry maps an edge to a set of
     *  edges intersecting with it. Execution simulation used here to collect
     *  buffer intersections is a universal but costly strategy.
     *
     *  @param graph The given SDF graph.
     *  @param schedule The given SDF schedule.
     *  @return The buffer intersections in the form of Map of Sets.
     */
    static public Map bufferIntersections(
            SDFGraph graph, SDFSchedule schedule) {

        Iterator history = bufferHistory(graph, schedule);

        // initialize the buffer intersection map
        HashMap intersectionMap = new HashMap();
        Iterator edges = graph.edges().iterator();
        while (edges.hasNext()) {
            intersectionMap.put((Edge)edges.next(), new HashSet());
        }

        // Go through the history and check all living buffers.
        while (history.hasNext()) {
            Set buffers = ((Map)history.next()).keySet();
            Iterator alives = buffers.iterator();
            while (alives.hasNext()) {
                Edge buffer = (Edge)alives.next();
                Set intersection = (Set)intersectionMap.get(buffer);
                intersection.addAll(buffers);
                // exclude the buffer itself
                intersection.remove(buffer);
            }
        }
        return intersectionMap;
    }

    /** Get the lower bound for buffer sharing cost. Sharing cost means the
     *  total memory space needed for shared buffers.
     *  @return The lower bound for buffer sharing cost.
     */
    public int bufferSharingLowerBound() {
        return bufferSharingLowerBound(
                (SDFGraph)graph(), (SDFSchedule)schedule());
    }

    /** Get the lower bound for buffer sharing cost. Sharing cost means the
     *  total memory space needed for shared buffers. This is a universal
     *  strategy for any schedules.
     *
     *  @param graph The given graph.
     *  @param schedule The given intersection map.
     *  @param size The buffer size array ordered by graph edge labels.
     *  @return The lower bound for buffer sharing cost.
     */
    static public int bufferSharingLowerBound(
            SDFGraph graph, SDFSchedule schedule) {

        Iterator history = bufferHistory(graph, schedule);
        int[] size = bufferSize(graph, schedule);

        // Go through buffer history and check
        // buffer cost at every instant.
        int bound = 0;
        while (history.hasNext()) {
            // Total buffer cost for this instant.
            int cost = 0;
            Iterator buffers = ((Map)history.next()).keySet().iterator();
            while (buffers.hasNext()) {
                int label = graph.edgeLabel((Edge)buffers.next());
                cost += size[label];
            }
            if (cost > bound)
                bound = cost;
        }
        return bound;
    }

    /** Size for individual buffers and sum of them. The last array
     *  element is the sum of all buffer size.
     *
     *  @return Size for individual buffers and sum of them.
     */
    public int[] bufferSize() {
        return bufferSize((SDFGraph)graph(), (SDFSchedule)schedule());
    }

    /** Compute buffer size for a given SDF graph and SDF schedule.
     *  Array elements from 0 to graph.edgeCount() - 1 are sizes for
     *  all buffers; element at graph.edgeCount() is the total size
     *  for whole graph.
     *  @param graph The given SDF graph.
     *  @param schedule The given SDF schedule.
     *  @return An array of sizez.
     */
    static public int[] bufferSize(SDFGraph graph, SDFSchedule schedule) {

        Iterator history = bufferHistory(graph, schedule);

        int totalBuffers = graph.edgeCount();
        int[] bufferSize = new int[totalBuffers + 1];

        // Go through buffer living history to check the maximum
        // tokens residing in each buffer.
        while (history.hasNext()) {
            Map bufferMap = (Map)history.next();
            Iterator buffers = bufferMap.keySet().iterator();
            while (buffers.hasNext()) {
                Edge buffer = (Edge)buffers.next();
                int label = graph.edgeLabel(buffer);
                int token = ((Integer)bufferMap.get(buffer)).intValue();
                if (token > bufferSize[label])
                    bufferSize[label] = token;
            }
        }

        // Sum up every buffer's size to figure out
        // total size needed for the whole graph
        int cost = 0;
        for (int i = 0; i < totalBuffers; i++)
            cost += bufferSize[i];
        bufferSize[totalBuffers] = cost;

        return bufferSize;
    }

    /** Test compatibility in terms of the class of graph. Only
     *  {@link SDFGraph}s is compatible with <code>SDFScheduler</code>.
     *
     *  @param graph The given graph to test compatibility.
     *  @return <code>True</code> if compatible; <code>false</code> otherwise.
     */
    public boolean compatible(Graph graph) {
        return (graph instanceof SDFGraph);
    }

    /** Return a incompatibility description in terms of the
     *  class of graph.
     *
     *  @param graph The given graph.
     *  @return A description of incompatibility.
     */
    public String incompatibilityDescription(Graph graph) {
        String result = new String();
        if (!(graph instanceof SDFGraph)) {
            result = "The given graph of class " + graph.getClass().getName() +
                     " is not an instance of SDFGraph";
        }
        return result;
    }

    /** A desrciption of the scheduler.
     *
     *  @return A text description.
     */
    public String toString() {
        return new String("A base SDFGraph scheduler.");
    }

    ///////////////////////////////////////////////////////////////////
    ////                       protected methods                   ////

    /** Convert a Ptolemy to {@link SDFSchedule}.
     *
     *  @param ptSchedule The Ptolemy II schedule.
     *  @param sdfSchedule The SDF schedule.
     */
    protected void _toSDFSchedule(
            ScheduleElement ptSchedule, SDFSchedule sdfSchedule) {

        sdfSchedule.setVectorizationFactor(ptSchedule.getIterationCount());

        // only one node in the original schedule
        if (ptSchedule instanceof Firing) {
            sdfSchedule.add(ptSchedule);
        // more than one node in the original graph
        } else {
            Iterator elements = ((Schedule)ptSchedule).iterator();
            while (elements.hasNext()) {
                sdfSchedule.add((ScheduleElement)elements.next());
            }
        }
    }

    ///////////////////////////////////////////////////////////////////
    ////                      protected variables                  ////

    /* The result {@link SDFSchedule}. */
    protected SDFSchedule _sdfSchedule;
}
